import numpy as np
import numpy.testing as npt
import pytest
try:
    # Python 3
    from unittest import mock
except ImportError:
    # Python 2
    import mock

try:
    # Python 3
    from imp import reload
except ImportError:
    pass

from .. import files
from .. import utils


def test_set_skvideo_path():
    # Smoke-test
    files.set_skvideo_path('/usr/bin')
    files.set_skvideo_path(libav_path='/usr/bin')


def test_load_video_metadata():
    # Load a test example
    reload(files)
    with pytest.raises(OSError):
        metadata = files.load_video_metadata('nothing_there.mp4')

    from skvideo import datasets
    metadata = files.load_video_metadata(datasets.bikes())
    npt.assert_equal(metadata['@codec_name'], 'h264')
    npt.assert_equal(metadata['@duration_ts'], '128000')
    npt.assert_equal(metadata['@r_frame_rate'], '25/1')

    # Trigger an import error
    with mock.patch.dict("sys.modules", {"skvideo": {}, "skvideo.utils": {}}):
        with pytest.raises(ImportError):
            reload(files)
            files.load_video_metadata(datasets.bikes())


def test_load_framerate():
    # Load a test example
    reload(files)
    with pytest.raises(OSError):
        files.load_video_metadata('nothing_there.mp4')

    from skvideo import datasets
    fps = files.load_video_framerate(datasets.bikes())
    npt.assert_equal(fps, 25)

    # Trigger an import error
    with mock.patch.dict("sys.modules", {"skvideo": {}, "skvideo.utils": {}}):
        with pytest.raises(ImportError):
            reload(files)
            files.load_video_framerate(datasets.bikes())


def test_load_video():
    reload(files)
    # Load a test example
    from skvideo import datasets

    # Load with default values
    video = files.load_video(datasets.bikes(), as_timeseries=False)
    npt.assert_equal(isinstance(video, np.ndarray), True)
    npt.assert_equal(video.shape, [250, 272, 640, 3])

    # Load as grayscale
    video = files.load_video(datasets.bikes(), as_timeseries=False,
                             as_gray=True)
    npt.assert_equal(isinstance(video, np.ndarray), True)
    npt.assert_equal(video.shape, [250, 272, 640, 1])

    # Load as TimeSeries
    video = files.load_video(datasets.bikes(), as_timeseries=True)
    fps = files.load_video_framerate(datasets.bikes())
    npt.assert_equal(isinstance(video, utils.TimeSeries), True)
    npt.assert_almost_equal(video.tsample, 1.0 / fps)
    npt.assert_equal(video.shape, [272, 640, 3, 250])

    # Trigger an import error
    with mock.patch.dict("sys.modules", {"skvideo": {}, "skvideo.utils": {}}):
        with pytest.raises(ImportError):
            reload(files)
            files.load_video('invalid.avi')


def test_load_video_generator():
    # Load a test example
    reload(files)

    from skvideo import datasets
    reader = files.load_video_generator(datasets.bikes())
    for frame in reader.nextFrame():
        npt.assert_equal(frame.shape, [272, 640, 3])

    # Trigger an import error
    with mock.patch.dict("sys.modules", {"skvideo": {}, "skvideo.utils": {}}):
        with pytest.raises(ImportError):
            reload(files)
            files.load_video_generator('invalid.avi')


def test_save_video():
    # Load a test example
    reload(files)
    from skvideo import datasets

    # There and back again: ndarray
    videoin = files.load_video(datasets.bikes(), as_timeseries=False)
    fpsin = files.load_video_framerate(datasets.bikes())
    files.save_video(videoin, 'myvideo.mp4', fps=fpsin)
    videout = files.load_video('myvideo.mp4', as_timeseries=False)
    npt.assert_equal(videoin.shape, videout.shape)
    npt.assert_almost_equal(videout / 255.0, videoin / 255.0, decimal=0)

    # Write to file with different frame rate, widths, and heights
    fpsout = 15
    files.save_video(videoin, 'myvideo.mp4', width=100, fps=fpsout)
    npt.assert_equal(files.load_video_framerate('myvideo.mp4'), fpsout)
    videout = files.load_video('myvideo.mp4', as_timeseries=False)
    npt.assert_equal(videout.shape[2], 100)
    files.save_video(videoin, 'myvideo.mp4', height=20, fps=fpsout)
    videout = files.load_video('myvideo.mp4', as_timeseries=False)
    npt.assert_equal(videout.shape[1], 20)
    videout = None

    # There and back again: TimeSeries
    tsamplein = 1.0 / float(fpsin)
    tsampleout = 1.0 / float(fpsout)
    rollaxes = np.roll(range(videoin.ndim), -1)
    tsin = utils.TimeSeries(tsamplein, np.transpose(videoin, rollaxes))
    files.save_video(tsin, 'myvideo.mp4', fps=fpsout)
    npt.assert_equal(tsin.tsample, tsamplein)
    tsout = files.load_video('myvideo.mp4', as_timeseries=True)
    npt.assert_equal(files.load_video_framerate('myvideo.mp4'), fpsout)
    npt.assert_equal(isinstance(tsout, utils.TimeSeries), True)
    npt.assert_almost_equal(tsout.tsample, tsampleout)

    # Also verify the actual data
    tsres = tsin.resample(tsampleout)
    npt.assert_equal(tsout.shape, tsres.shape)
    npt.assert_almost_equal(tsout.data / 255.0, tsres.data / tsres.data.max(),
                            decimal=0)

    with pytest.raises(TypeError):
        files.save_video([2, 3, 4], 'invalid.avi')

    # Trigger an import error
    with mock.patch.dict("sys.modules", {"skvideo": {}}):
        with pytest.raises(ImportError):
            reload(files)
            files.save_video(videoin, 'invalid.avi')
    with mock.patch.dict("sys.modules", {"skimage": {}}):
        with pytest.raises(ImportError):
            reload(files)
            files.save_video(videoin, 'invalid.avi')


def test_save_video_sidebyside():
    reload(files)
    from skvideo import datasets
    videoin = files.load_video(datasets.bikes(), as_timeseries=False)
    fps = files.load_video_framerate(datasets.bikes())
    tsample = 1.0 / float(fps)
    rollaxes = np.roll(range(videoin.ndim), -1)
    percept = utils.TimeSeries(tsample, np.transpose(videoin, rollaxes))

    files.save_video_sidebyside(datasets.bikes(), percept, 'mymovie.mp4',
                                fps=fps)
    videout = files.load_video('mymovie.mp4', as_timeseries=False)
    npt.assert_equal(videout.shape[0], videoin.shape[0])
    npt.assert_equal(videout.shape[1], videoin.shape[1])
    npt.assert_equal(videout.shape[2], videoin.shape[2] * 2)
    npt.assert_equal(videout.shape[3], videoin.shape[3])

    with pytest.raises(TypeError):
        files.save_video_sidebyside(datasets.bikes(), [2, 3, 4], 'invalid.avi')

    with mock.patch.dict("sys.modules", {"skvideo": {}}):
        with pytest.raises(ImportError):
            reload(files)
            files.save_video_sidebyside(datasets.bikes(), percept,
                                        'invalid.avi')
    with mock.patch.dict("sys.modules", {"skimage": {}}):
        with pytest.raises(ImportError):
            reload(files)
            files.save_video_sidebyside(datasets.bikes(), percept,
                                        'invalid.avi')


def test_find_files_like():
    # Not sure how to check a valid pattern match, because we
    # don't know in which directory the test suite is
    # executed... so smoke test
    files.find_files_like(".", ".*")

    # Or fail to find an unlikely file name
    filenames = files.find_files_like(".", "theresnowaythisexists\.xyz")
    npt.assert_equal(filenames, [])



import numpy as np

import scipy.special as ss
import scipy.spatial as spat
import abc
import six
import os.path
import logging

from . import fast_retina as fr
from . import utils


SUPPORTED_LAYERS = ['INL', 'GCL', 'OFL']
SUPPORTED_TEMPORAL_MODELS = ['latest', 'Nanduri2012', 'Horsager2009']


class Grid(object):
    """Represent the retinal coordinate frame"""

    def __init__(self, x_range=(-1000.0, 1000.0), y_range=(-1000.0, 1000.0),
                 eye='RE', sampling=25, n_axons=501, phi_range=(-180.0, 180.0),
                 n_rho=801, rho_range=(4.0, 45.0), loc_od=(15.5, 1.5),
                 sensitivity_rule='decay', contribution_rule='max',
                 decay_const=2.0, powermean_exp=None, datapath='.',
                 save_data=True, engine='joblib', scheduler='threading',
                 n_jobs=-1):
        """Generates a spatial grid representing the retinal coordinate frame

        This function generates the coordinate system for the retina
        and an axon map. As this can take a while, the function will
        first look for an already existing file in the directory `datapath`
        that was automatically created from an earlier call to this function,
        before it attempts to generate new grid from scratch.

        Parameters
        ----------
        x_range : (xlo, xhi), optional, default: xlo=-1000, xhi=1000
           Extent of the retinal coverage (microns) in horizontal dimension.
        y_range : (ylo, yhi), optional, default: ylo=-1000, ylo=1000
           Extent of the retinal coverage (microns) in vertical dimension.
        eye : {'LE', 'RE'}, optional, default: 'RE'
            Which eye to simulate (left/right). The optic disc is at (15, 2)
            deg in a right eye, and at (-15, 2) deg in a left eye.
        sampling : float, optional, default: 25
            Spatial sampling step (microns) for the grid.
        n_axons : int, optional, default: 501
            The number of axons to generate. Their start orientations `phi0`
            (in modified polar coordinates) will be sampled uniformly from
            `phi_range`.
        phi_range : (lophi, hiphi), optional, default: (-180, 180)
            Range of angular positions of axon fibers at their starting points
            (polar coordinates, degrees) to be sampled uniformly with `n_axons`
            samples. Must be within [-180, 180].
        n_rho: int, optional, default: 801
            Number of sampling points along the radial axis(polar coordinates).
        rho_range: (rho_min, rho_max), optional, default: (4.0, 45.0)
            Lower and upper bounds for the radial position values(polar
            coordinates).
        loc_od: (x_od, y_od), optional, default: (15.0, 2.0)
            Location of the center of the optic disc(x, y) in Cartesian
            coordinates.
        sensitivity_rule : {'decay', 'Jeng2011'}, optional, default: 'decay'
            This rule specifies how the activation of the axon differs as a
            function of distance from the soma. The following options are
            available:
            - 'decay':
                Axon sensitivity decays exponentially with distance. Specify
                `decay_const` to change the steepness of the fall-off with
                distance.
            - 'Jeng2011':
                Axon sensitivity peaks near the sodium band (50um from the
                soma), then plateaus on the distal axon at roughly half of the
                peak sensitivity. See Figure 2 in Jeng, Tang, Molnar, Desai,
                and Fried (2011). The sodium channel band shapes the response
                to electric stimulation in retinal ganglion cells. J Neural Eng
                8 (036022).
        contribution_rule : {'max', 'sum', 'mean'}, optional, default: 'max'
            This rule specifies how the activation thresholds across all axon
            segments are combined to determine the contribution of the axon to
            the current spread. The following options are available:
            - 'max':
                The axon's contribution to the current spread is equal to the
                max. sensitivity across all axon segments.
            - 'sum':
                The axon's contribution to the current spread is equal to the
                sum sensitivity across all axon segments.
            - 'mean':
                The axon's contribution to the current spread is equal to the
                mean sensitivity across all axon segments. Specify
                `powermean_exp` to change the exponent of the generalized
                (power) mean, calculated as np.mean(x ** powermean_exp) **
                (1.0 / powermean_exp).
        decay_const : float, optional, default: 2.0
            When `sensitivity_rule` is set to 'decay', specifies the decay
            constant of the exponential fall-off.
        powermean_exp : float, optional, default: None
            When `sensitivity_rule` is set to 'mean', specifies the exponent of
            the generalized (power) mean function. The power mean is calculated
            as np.mean(x ** powermean_exp) ** (1.0 / powermean_exp).
        datapath : str, optional, default: current directory
            Relative path where to look for existing retina files, and where to
            store new files.
        save_data : bool, optional, default: True
            Flag whether to save the data to a new file (True) or not (False).
            The file name is automatically generated from all specified input
            arguments.
        engine : str, optional, default: 'joblib'
            Which computational back end to use:
            - 'serial': Single-core computation
            - 'joblib': Parallelization via joblib (requires `pip install
                        joblib`)
            - 'dask': Parallelization via dask (requires `pip install dask`).
                      Dask backend can be specified via `threading`.
        scheduler : str, optional, default: 'threading'
            Which scheduler to use (irrelevant for 'serial' engine):
            - 'threading': a scheduler backed by a thread pool
            - 'multiprocessing': a scheduler backed by a process pool
        n_jobs : int, optional, default: -1
            Number of cores (threads) to run the model on in parallel.
            Specify -1 to use as many cores as available.
        """
        if not isinstance(x_range, (tuple, list, np.ndarray)):
            raise ValueError('`x_range` must be a tuple (`xlo`, `xhi`).')
        if x_range[0] > x_range[1]:
            raise ValueError('Lower bound on x cannot be larger than the '
                             'upper bound.')
        if not isinstance(y_range, (tuple, list, np.ndarray)):
            raise ValueError('`y_range` must be a tuple (`ylo`, `yhi`).')
        if y_range[0] > y_range[1]:
            raise ValueError('Lower bound on y cannot be larger than the '
                             'upper bound.')
        if n_axons < 1:
            raise ValueError('Number of axons must be >= 1.')
        if np.any(np.abs(phi_range) > 180.0):
            raise ValueError('phi must be within [-180, 180].')
        if phi_range[0] > phi_range[1]:
            raise ValueError('Lower bound on phi cannot be larger than the '
                             'upper bound.')
        self.x_range = x_range
        self.y_range = y_range
        self.sampling = sampling
        self.sensitivity_rule = sensitivity_rule
        self.contribution_rule = contribution_rule
        self.decay_const = decay_const
        self.powermean_exp = powermean_exp
        self.engine = engine
        self.scheduler = scheduler
        self.n_jobs = n_jobs

        if np.abs(loc_od[1]) > 10.0:
            logging.getLogger(__name__).warn("The Jansonius model might "
                                             "misbehave if `loc_od` has "
                                             "a y value > 10.")

        if eye.upper() == 'RE':
            if loc_od[0] <= 0:
                w_s = "In a right eye, the optic disc usually has x > 0 - "
                w_s += "currently at (%.1f, %.1f)." % (loc_od[0], loc_od[1])
                logging.getLogger(__name__).warn(w_s)
        elif eye.upper() == 'LE':
            if loc_od[0] > 0:
                w_s = "In a left eye, the optic disc usually has x < 0 - "
                w_s += "currently at (%.1f, %.1f)." % (loc_od[0], loc_od[1])
                logging.getLogger(__name__).warn(w_s)
        else:
            e_s = "Unknown eye string '%s': Choose from 'LE', 'RE'." % eye
            raise ValueError(e_s)

        # Include endpoints in meshgrid
        xlo, xhi = x_range
        ylo, yhi = y_range
        num_x = int((xhi - xlo) / sampling + 1)
        num_y = int((yhi - ylo) / sampling + 1)
        self.gridx, self.gridy = np.meshgrid(np.linspace(xlo, xhi, num_x),
                                             np.linspace(ylo, yhi, num_y),
                                             indexing='xy')

        # Create descriptive filename based on input args
        filename = "retina_%s_s%d_a%d_r%d_%dx%d.npz" % (eye, sampling, n_axons,
                                                        n_rho, xhi - xlo,
                                                        yhi - ylo)
        filename = os.path.join(datapath, filename)

        # There are some variables, like `sensitivity_rule` and `decay_const`
        # that are only needed in the effective current calculation, not for
        # the grid and axon maps - so not included here:
        grid_dict = {'x_range': x_range, 'y_range': y_range, 'eye': eye,
                     'n_axons': n_axons, 'phi_range': phi_range,
                     'n_rho': n_rho, 'rho_range': rho_range,
                     'sampling': sampling, 'loc_od': loc_od}

        # Assign all elements in the dictionary to this object
        for key, value in six.iteritems(grid_dict):
            setattr(self, key, value)

        # Check if such a file already exists. If so, load parameters and
        # make sure they are the same as specified above. Else, create new.
        need_new_grid = False
        if not os.path.exists(filename):
            need_new_grid = True
        else:
            logging.getLogger(__name__).info('Loading file "%s".' % filename)
            try:
                load_grid_dict = six.moves.cPickle.load(open(filename, 'rb'))
            except six.moves.cPickle.UnpicklingError:
                msg = 'UnpicklingError: Could not load file "%s".'
                logging.getLogger(__name__).info(msg)
                need_new_grid = True
                load_grid_dict = {}

            # Make sure all relevant variables are present and have the right
            # values:
            for key, value in six.iteritems(grid_dict):
                if key not in load_grid_dict:
                    logging.getLogger(__name__).info('File out of date.')
                    need_new_grid = True
                    break
                if value != load_grid_dict[key]:
                    logging.getLogger(__name__).info('File out of date.')
                    need_new_grid = True
                    break

        # At this point we know whether we need to generate a new retina:
        if need_new_grid:
            info_str = "Generating new file '%s'." % filename
            logging.getLogger(__name__).info(info_str)

            # Grow a number `n_axons` of axon bundles with orientations in
            # `phi_range`
            phi = np.linspace(*phi_range, num=n_axons)
            func_kwargs = {'n_rho': n_rho, 'rho_range': rho_range,
                           'loc_od': loc_od, 'eye': eye}
            self.axon_bundles = utils.parfor(jansonius2009, phi,
                                             func_kwargs=func_kwargs,
                                             engine=engine, n_jobs=n_jobs,
                                             scheduler=scheduler)
            grid_dict['axon_bundles'] = self.axon_bundles

            # Assume there is a neuron at every grid location: Use the above
            # axon bundles to assign an axon to each neuron
            xg, yg = ret2dva(self.gridx), ret2dva(self.gridy)
            pos_xy = np.column_stack((xg.ravel(), yg.ravel()))
            self.axons = utils.parfor(find_closest_axon, pos_xy,
                                      func_args=[self.axon_bundles])
            grid_dict['axons'] = self.axons

            # For every axon segment, calculate distance to soma. Snap axon
            # locations to the grid using a nearest-neighbor tree structure:
            func_kwargs = {'tree': spat.cKDTree(pos_xy)}
            self.axon_distances = utils.parfor(axon_dist_from_soma, self.axons,
                                               func_args=[xg, yg],
                                               func_kwargs=func_kwargs,
                                               engine=engine, n_jobs=n_jobs,
                                               scheduler=scheduler)
            grid_dict['axon_distances'] = self.axon_distances

            # Save the variables, together with metadata about the grid:
            if save_data:
                six.moves.cPickle.dump(grid_dict, open(filename, 'wb'))
        else:
            # Assign all elements in the loaded dictionary to this object
            for key, value in six.iteritems(load_grid_dict):
                setattr(self, key, value)

    def current2effectivecurrent(self, current_spread):
        """

        Converts a current spread map to an 'effective' current spread map, by
        passing the map through a mapping of axon streaks.

        Parameters
        ----------
        cs: array
            The 2D spread map in retinal space

        Returns
        -------
        ecm: array
            The effective current spread, a time series of the same size as
            the current map, where each pixel is the dot product of the pixel
            values in ecm along the pixels in the list in axon_map, weighted
            by the weights axon map.
        """
        contrib = utils.parfor(axon_contribution, self.axon_distances,
                               func_args=[current_spread], engine=self.engine,
                               func_kwargs={
                                   'sensitivity_rule': self.sensitivity_rule,
                                   'contribution_rule': self.contribution_rule,
                                   'decay_const': self.decay_const,
                                   'powermean_exp': self.powermean_exp
                               },
                               scheduler=self.scheduler, n_jobs=self.n_jobs)

        ecs = np.zeros_like(current_spread)
        px_contrib = list(filter(None, contrib))
        for idx, value in px_contrib:
            ecs.ravel()[idx] = value

        # Normalize so that the max of `ecs` is the same as `current_spread`
        return ecs / (ecs.max() + np.finfo(float).eps) * current_spread.max()

    def electrode_ecs(self, implant, alpha=14000, n=1.69):
        """
        Gather current spread and effective current spread for each electrode
        within both the bipolar and the ganglion cell layer

        Parameters
        ----------
        implant: implants.ElectrodeArray
            An implants.ElectrodeArray instance describing the implant.

        alpha: float
            Current spread parameter
        n: float
            Current spread parameter

        Returns
        -------
        ecs: contains n arrays containing the the effective current
            spread within various layers
            for each electrode in the array respectively.

        See also
        --------
        Electrode.current_spread
        """

        cs = np.zeros((self.gridx.shape[0], self.gridx.shape[1],
                       2, len(implant.electrodes)), dtype=float)
        ecs = np.zeros((self.gridx.shape[0], self.gridx.shape[1],
                        2, len(implant.electrodes)), dtype=float)

        for i, e in enumerate(implant.electrodes):
            cs[..., 0, i] = e.current_spread(self.gridx, self.gridy,
                                             layer='INL', alpha=alpha, n=n)
            ecs[..., 0, i] = cs[..., 0, i]
            cs[..., 1, i] = e.current_spread(self.gridx, self.gridy,
                                             layer='OFL', alpha=alpha, n=n)
            ecs[:, :, 1, i] = self.current2effectivecurrent(cs[..., 1, i])

        return ecs, cs


@six.add_metaclass(abc.ABCMeta)
class BaseModel():
    """Abstract base class for all models of temporal sensitivity.

    This class provides a standard template for all models of temporal
    sensitivity.
    """

    def set_kwargs(self, warn_inexistent, **kwargs):
        """Overwrite any given keyword arguments

        Parameters
        ----------
        warn_inexistent: bool
            If True, displays a warning message if a keyword is provided that
            is not recognized by the temporal model.
        """
        for key, value in six.iteritems(kwargs):
            if not hasattr(self, key) and warn_inexistent:
                w_s = "Unknown class attribute '%s'" % key
                logging.getLogger(__name__).warning(w_s)
            setattr(self, key, value)

    def __init__(self, **kwargs):
        self.set_kwargs(True, **kwargs)

    @abc.abstractmethod
    def model_cascade(self, in_arr, pt_list, layers, use_jit):
        """Abstract base ganglion cell model

        Parameters
        ----------
        in_arr: array - like
            A 2D array specifying the effective current values at a particular
            spatial location(pixel); one value per retinal layer and
            electrode. Dimensions: <  # layers x #electrodes>
        pt_list: list
            List of pulse train 'data' containers.
            Dimensions: <  # electrodes x #time points>
        layers: list
            List of retinal layers to simulate.
            Choose from:
            - 'OFL': optic fiber layer
            - 'GCL': ganglion cell layer
            - 'INL': inner nuclear layer
        use_jit: bool
            If True, applies just - in-time(JIT) compilation to expensive
            computations for additional speed - up(requires Numba).
        """
        pass

    # Static attribute
    tsample = 0.005 / 1000


class Horsager2009(BaseModel):
    """Model of temporal sensitivity (Horsager et al. 2009)

    This class implements the model of temporal sensitivty as described in:
    > A Horsager, SH Greenwald, JD Weiland, MS Humayun, RJ Greenberg,
    > MJ McMahon, GM Boynton, and I Fine(2009). Predicting visual sensitivity
    > in retinal prosthesis patients. Investigative Ophthalmology & Visual
    > Science, 50(4): 1483.

    Parameters
    ----------
    tsample: float, optional, default: 0.005 / 1000 seconds
        Sampling time step(seconds).
    tau1: float, optional, default: 0.42 / 1000 seconds
        Time decay constant for the fast leaky integrater of the ganglion
        cell layer(GCL).
    tau2: float, optional, default: 45.25 / 1000 seconds
        Time decay constant for the charge accumulation, has values
        between 38 - 57 ms.
    tau3: float, optional, default: 26.25 / 1000 seconds
        Time decay constant for the slow leaky integrator.
        Default: 26.25 / 1000 s.
    epsilon: float, optional, default: 8.73
        Scaling factor applied to charge accumulation(used to be called
        epsilon).
    beta: float, optional, default: 3.43
        Power nonlinearity applied after half - rectification. The original model
        used two different values, depending on whether an experiment is at
        threshold(`beta`=3.43) or above threshold(`beta`=0.83).
    """

    def __init__(self, **kwargs):
        self.tsample = 0.01 / 1000
        self.tau1 = 0.42 / 1000
        self.tau2 = 45.25 / 1000
        self.tau3 = 26.25 / 1000
        self.epsilon = 2.25
        self.beta = 3.43

        # Overwrite any given keyword arguments, print warning message (True)
        # if attempting to set an unrecognized keyword
        self.set_kwargs(True, **kwargs)

        _, self.gamma1 = utils.gamma(1, self.tau1, self.tsample)
        _, self.gamma2 = utils.gamma(1, self.tau2, self.tsample)
        _, self.gamma3 = utils.gamma(3, self.tau3, self.tsample)

    def calc_layer_current(self, in_arr, pt_list, layers):
        """Calculates the effective current map of a given layer

        Parameters
        ----------
        in_arr: array - like
            A 2D array specifying the effective current values
            at a particular spatial location(pixel); one value
            per retinal layer and electrode.
            Dimensions: <  # layers x #electrodes>
        pt_list: list
            List of pulse train 'data' containers.
            Dimensions: <  # electrodes x #time points>
        layers: list
            List of retinal layers to simulate.
            Choose from:
            - 'OFL': optic fiber layer
            - 'GCL': ganglion cell layer
        """
        if 'INL' in layers:
            raise ValueError("The Horsager2009 model does not support an "
                             "inner nuclear layer.")

        if ('GCL' or 'OFL') in layers:
            ecm = np.sum(in_arr[1, :, np.newaxis] * pt_list, axis=0)
        else:
            raise ValueError("Acceptable values for `layers` are: 'GCL', "
                             "'OFL'.")
        return ecm

    def model_cascade(self, in_arr, pt_list, layers, use_jit):
        """Horsager model cascade

        Parameters
        ----------
        in_arr: array - like
            A 2D array specifying the effective current values
            at a particular spatial location(pixel); one value
            per retinal layer and electrode.
            Dimensions: <  # layers x #electrodes>
        pt_list: list
            List of pulse train 'data' containers.
            Dimensions: <  # electrodes x #time points>
        layers: list
            List of retinal layers to simulate.
            Choose from:
            - 'OFL': optic fiber layer
            - 'GCL': ganglion cell layer
        use_jit: bool
            If True, applies just - in-time(JIT) compilation to
            expensive computations for additional speed - up
            (requires Numba).
        """
        if 'INL' in layers:
            raise ValueError("The Nanduri2012 model does not support an inner "
                             "nuclear layer.")

        # Although the paper says to use cathodic-first, the code only
        # reproduces if we use what we now call anodic-first. So flip the sign
        # on the stimulus here:
        stim = -self.calc_layer_current(in_arr, pt_list, layers)

        # R1 convolved the entire stimulus (with both pos + neg parts)
        r1 = self.tsample * utils.conv(stim, self.gamma1, mode='full',
                                       method='sparse')[:stim.size]

        # It's possible that charge accumulation was done on the anodic phase.
        # It might not matter too much (timing is slightly different, but the
        # data are not accurate enough to warrant using one over the other).
        # Thus use what makes the most sense: accumulate on cathodic
        ca = self.tsample * np.cumsum(np.maximum(0, -stim))
        ca = self.tsample * utils.conv(ca, self.gamma2, mode='full',
                                       method='fft')[:stim.size]
        r2 = r1 - self.epsilon * ca

        # Then half-rectify and pass through the power-nonlinearity
        r3 = np.maximum(0.0, r2) ** self.beta

        # Then convolve with slow gamma
        r4 = self.tsample * utils.conv(r3, self.gamma3, mode='full',
                                       method='fft')[:stim.size]

        return utils.TimeSeries(self.tsample, r4)


class Nanduri2012(BaseModel):
    """Model of temporal sensitivity (Nanduri et al. 2012)

    This class implements the model of temporal sensitivity as described in:
    > Nanduri, Fine, Horsager, Boynton, Humayun, Greenberg, Weiland(2012).
    > Frequency and Amplitude Modulation Have Different Effects on the Percepts
    > Elicited by Retinal Stimulation. Investigative Ophthalmology & Visual
    > Science January 2012, Vol.53, 205 - 214. doi: 10.1167 / iovs.11 - 8401.

    Parameters
    ----------
    tsample: float, optional, default: 0.005 / 1000 seconds
        Sampling time step(seconds).
    tau1: float, optional, default: 0.42 / 1000 seconds
        Time decay constant for the fast leaky integrater of the ganglion
        cell layer(GCL).
    tau2: float, optional, default: 45.25 / 1000 seconds
        Time decay constant for the charge accumulation, has values
        between 38 - 57 ms.
    tau3: float, optional, default: 26.25 / 1000 seconds
        Time decay constant for the slow leaky integrator.
        Default: 26.25 / 1000 s.
    eps: float, optional, default: 8.73
        Scaling factor applied to charge accumulation(used to be called
        epsilon).
    asymptote: float, optional, default: 14.0
        Asymptote of the logistic function used in the stationary
        nonlinearity stage.
    slope: float, optional, default: 3.0
        Slope of the logistic function in the stationary nonlinearity
        stage.
    shift: float, optional, default: 16.0
        Shift of the logistic function in the stationary nonlinearity
        stage.
    """

    def __init__(self, **kwargs):
        # Set default values of keyword arguments
        self.tau1 = 0.42 / 1000
        self.tau2 = 45.25 / 1000
        self.tau3 = 26.25 / 1000
        self.eps = 8.73
        self.asymptote = 14.0
        self.slope = 3.0
        self.shift = 16.0

        # Nanduri (2012) has a term in the stationary nonlinearity step that
        # depends on future values of R3: max_t(R3). Because the finite
        # difference model cannot look into the future, we need to set a
        # scaling factor here:
        self.maxR3 = 100.0

        # Overwrite any given keyword arguments, print warning message (True)
        # if attempting to set an unrecognized keyword
        self.set_kwargs(True, **kwargs)

    def calc_layer_current(self, in_arr, pt_list):
        """Calculates the effective current map of a given layer

        Parameters
        ----------
        in_arr: array - like
            A 2D array specifying the effective current values
            at a particular spatial location(pixel); one value
            per retinal layer and electrode.
            Dimensions: <  # layers x #electrodes>
        pt_list: list
            List of pulse train 'data' containers.
            Dimensions: <  # electrodes x #time points>
        """
        in_flat = in_arr[1, :].reshape(-1).astype(float)
        pt_arr = np.array(pt_list, dtype=float)
        return np.array(fr.nanduri2012_calc_layer_current(in_flat, pt_arr))

    def model_cascade(self, in_arr, pt_list, layers, use_jit):
        """Nanduri model cascade

        Parameters
        ----------
        in_arr: array - like
            A 2D array specifying the effective current values
            at a particular spatial location(pixel); one value
            per retinal layer and electrode.
            Dimensions: < #layers x #electrodes>
        pt_list: list
            List of pulse train 'data' containers.
            Dimensions: < #electrodes < #time points > >
        layers: list
            List of retinal layers to simulate.
            Choose from:
            - 'OFL': optic fiber layer
            - 'GCL': ganglion cell layer
        use_jit: bool
            If True, applies just - in-time(JIT) compilation to
            expensive computations for additional speed - up
            (requires Numba).
        """
        if 'INL' in layers:
            raise ValueError("The Nanduri2012 model does not support an inner "
                             "nuclear layer.")
        if 'GCL' not in layers and 'OFL' not in layers:
            raise ValueError("Acceptable values for `layers` are: 'GCL', "
                             "'OFL'.")

        pulse = self.calc_layer_current(in_arr, pt_list)
        percept = fr.nanduri2012_model_cascade(pulse, self.tsample,
                                               self.tau1, self.tau2, self.tau3,
                                               self.asymptote, self.shift,
                                               self.slope, self.eps,
                                               self.maxR3)
        return utils.TimeSeries(self.tsample, percept)


class TemporalModel(BaseModel):
    """Latest edition of the temporal sensitivity model (experimental)

    This class implements the latest version of the temporal sensitivity
    model(experimental). As such, the model might still change from version
    to version. For more stable implementations, please refer to other,
    published models(see `p2p.retina.SUPPORTED_TEMPORAL_MODELS`).

    Parameters
    ----------
    tsample: float, optional, default: 0.005 / 1000 seconds
        Sampling time step(seconds).
    tau_gcl: float, optional, default: 45.25 / 1000 seconds
        Time decay constant for the fast leaky integrater of the ganglion
        cell layer(GCL).
        This is only important in combination with epiretinal electrode
        arrays.
    tau_inl: float, optional, default: 18.0 / 1000 seconds
        Time decay constant for the fast leaky integrater of the inner
        nuclear layer(INL); i.e., bipolar cell layer.
        This is only important in combination with subretinal electrode
        arrays.
    tau_ca: float, optional, default: 45.25 / 1000 seconds
        Time decay constant for the charge accumulation, has values
        between 38 - 57 ms.
    scale_ca: float, optional, default: 42.1
        Scaling factor applied to charge accumulation(used to be called
        epsilon).
    tau_slow: float, optional, default: 26.25 / 1000 seconds
        Time decay constant for the slow leaky integrator.
    scale_slow: float, optional, default: 1150.0
        Scaling factor applied to the output of the cascade, to make
        output values interpretable brightness values >= 0.
    lweight: float, optional, default: 0.636
        Relative weight applied to responses from bipolar cells(weight
        of ganglion cells is 1).
    aweight: float, optional, default: 0.5
        Relative weight applied to anodic charges(weight of cathodic
        charges is 1).
    slope: float, optional, default: 3.0
        Slope of the logistic function in the stationary nonlinearity
        stage.
    shift: float, optional, default: 15.0
        Shift of the logistic function in the stationary nonlinearity
        stage.
    """

    def __init__(self, **kwargs):
        # Set default values of keyword arguments
        self.tau_gcl = 0.42 / 1000
        self.tau_inl = 18.0 / 1000
        self.tau_ca = 45.25 / 1000
        self.tau_slow = 26.25 / 1000
        self.scale_ca = 42.1
        self.scale_slow = 1150.0
        self.lweight = 0.636
        self.aweight = 0.5
        self.slope = 3.0
        self.shift = 15.0

        # Overwrite any given keyword arguments, print warning message (True)
        # if attempting to set an unrecognized keyword
        self.set_kwargs(True, **kwargs)

        # perform one-time setup calculations
        _, self.gamma_inl = utils.gamma(1, self.tau_inl, self.tsample)
        _, self.gamma_gcl = utils.gamma(1, self.tau_gcl, self.tsample)

        # gamma_ca is used to calculate charge accumulation
        _, self.gamma_ca = utils.gamma(1, self.tau_ca, self.tsample)

        # gamma_slow is used to calculate the slow response
        _, self.gamma_slow = utils.gamma(3, self.tau_slow, self.tsample)

    def fast_response(self, stim, gamma, method, use_jit=True):
        """Fast response function

        Convolve a stimulus `stim` with a temporal low - pass filter `gamma`.

        Parameters
        ----------
        stim: array
           Temporal signal to process, stim(r, t) in Nanduri et al. (2012).
        use_jit: bool, optional
           If True (default), use numba just - in-time compilation.
        usefft: bool, optional
           If False (default), use sparseconv, else fftconvolve.

        Returns
        -------
        Fast response, b2(r, t) in Nanduri et al. (2012).

        Notes
        -----
        The function utils.sparseconv can be much faster than np.convolve and
        signal.fftconvolve if `stim` is sparse and much longer than the
        convolution kernel.
        The output is not converted to a TimeSeries object for speedup.
        """
        conv = utils.conv(stim, gamma, mode='full', method=method,
                          use_jit=use_jit)

        # Cut off the tail of the convolution to make the output signal
        # match the dimensions of the input signal.
        return self.tsample * conv[:stim.shape[-1]]

    def charge_accumulation(self, ecm):
        """Calculates the charge accumulation

        Charge accumulation is calculated on the effective input current
        `ecm`, as opposed to the output of the fast response stage.

        Parameters
        ----------
        ecm: array - like
            A 2D array specifying the effective current values at a particular
            spatial location(pixel); one value per retinal layer, averaged
            over all electrodes through that pixel.
            Dimensions: <  # layers x #time points>
        """
        ca = np.zeros_like(ecm)

        for i in range(ca.shape[0]):
            summed = self.tsample * np.cumsum(np.abs(ecm[i, :]))
            conved = self.tsample * utils.conv(summed, self.gamma_ca,
                                               mode='full', method='fft')
            ca[i, :] = self.scale_ca * conved[:ecm.shape[-1]]
        return ca

    def stationary_nonlinearity(self, stim):
        """Stationary nonlinearity

        Nonlinearly rescale a temporal signal `stim` across space and time,
        based on a sigmoidal function dependent on the maximum value of `stim`.
        This is Box 4 in Nanduri et al. (2012).
        The parameter values of the asymptote, slope, and shift of the logistic
        function are given by self.asymptote, self.slope, and self.shift,
        respectively.

        Parameters
        ----------
        stim: array
           Temporal signal to process, stim(r, t) in Nanduri et al. (2012).

        Returns
        -------
        Rescaled signal, b4(r, t) in Nanduri et al. (2012).

        Notes
        -----
        Conversion to TimeSeries is avoided for the sake of speedup.
        """
        # use expit (logistic) function for speedup
        sigmoid = ss.expit((stim.max() - self.shift) / self.slope)
        return stim * sigmoid

    def slow_response(self, stim):
        """Slow response function

        Convolve a stimulus `stim` with a low - pass filter(3 - stage gamma)
        with time constant self.tau_slow.
        This is Box 5 in Nanduri et al. (2012).

        Parameters
        ----------
        stim: array
           Temporal signal to process, stim(r, t) in Nanduri et al. (2012)

        Returns
        -------
        Slow response, b5(r, t) in Nanduri et al. (2012).

        Notes
        -----
        This is by far the most computationally involved part of the perceptual
        sensitivity model.
        Conversion to TimeSeries is avoided for the sake of speedup.
        """
        # No need to zero-pad: fftconvolve already takes care of optimal
        # kernel/data size
        conv = utils.conv(stim, self.gamma_slow, method='fft', mode='full')

        # Cut off the tail of the convolution to make the output signal match
        # the dimensions of the input signal.
        return self.scale_slow * self.tsample * conv[:stim.shape[-1]]

    def calc_layer_current(self, ecs_item, pt_list, layers):
        """For a given pixel, calculates the effective current for each retinal
           layer over time

        This function operates at a single - pixel level: It calculates the
        combined current from all electrodes through a spatial location
        over time. This calculation is performed per retinal layer.

        Parameters
        ----------
        ecs_item: array - like
            A 2D array specifying the effective current values at a
            particular spatial location(pixel); one value per retinal
            layer and electrode.
            Dimensions: <  # layers x #electrodes>
        pt_list: list
            A list of PulseTrain `data` containers.
            Dimensions: <  # electrodes x #time points>
        layers: list
            List of retinal layers to simulate. Choose from:
            - 'OFL': optic fiber layer
            - 'GCL': ganglion cell layer
            - 'INL': inner nuclear layer
        """
        not_supported = np.array([l not in SUPPORTED_LAYERS for l in layers],
                                 dtype=bool)
        if any(not_supported):
            raise ValueError("Acceptable values for `layers` is 'OFL', 'GCL', "
                             "'INL'.")

        ecm = np.zeros((ecs_item.shape[0], pt_list[0].shape[-1]))
        if 'INL' in layers:
            ecm[0, :] = np.sum(ecs_item[0, :, np.newaxis] * pt_list, axis=0)
        if ('GCL' or 'OFL') in layers:
            ecm[1, :] = np.sum(ecs_item[1, :, np.newaxis] * pt_list, axis=0)
        return ecm

    def model_cascade(self, ecs_item, pt_list, layers, use_jit):
        """The Temporal Sensitivity model

        This function applies the model of temporal sensitivity to a single
        retinal cell(i.e., a pixel). The model is inspired by Nanduri
        et al. (2012), with some extended functionality.

        Parameters
        ----------
        ecs_item: array - like
            A 2D array specifying the effective current values at a particular
            spatial location(pixel); one value per retinal layer and
            electrode.
            Dimensions: <  # layers x #electrodes>
        pt_list: list
            A list of PulseTrain `data` containers.
            Dimensions: <  # electrodes x #time points>
        layers: list
            List of retinal layers to simulate. Choose from:
            - 'OFL': optic fiber layer
            - 'GCL': ganglion cell layer
            - 'INL': inner nuclear layer
        use_jit: bool
            If True, applies just - in-time(JIT) compilation to expensive
            computations for additional speed - up(requires Numba).

        Returns
        -------
        Brightness response over time. In Nanduri et al. (2012), the
        maximum value of this signal was used to represent the perceptual
        brightness of a particular location in space, B(r).
        """
        # For each layer in the model, scale the pulse train data with the
        # effective current:
        ecm = self.calc_layer_current(ecs_item, pt_list, layers)

        # Calculate charge accumulation on the input
        ca = self.charge_accumulation(ecm)

        # Sparse convolution is faster if input is sparse. This is true for
        # the first convolution in the cascade, but not for subsequent ones.
        if 'INL' in layers:
            fr_inl = self.fast_response(ecm[0], self.gamma_inl,
                                        use_jit=use_jit,
                                        method='sparse')

            # Cathodic and anodic parts are treated separately: They have the
            # same charge accumulation, but anodic currents contribute less to
            # the response
            fr_inl_cath = np.maximum(0, -fr_inl)
            fr_inl_anod = self.aweight * np.maximum(0, fr_inl)
            resp_inl = np.maximum(0, fr_inl_cath + fr_inl_anod - ca[0, :])
        else:
            resp_inl = np.zeros_like(ecm[0])

        if ('GCL' or 'OFL') in layers:
            fr_gcl = self.fast_response(ecm[1], self.gamma_gcl,
                                        use_jit=use_jit,
                                        method='sparse')

            # Cathodic and anodic parts are treated separately: They have the
            # same charge accumulation, but anodic currents contribute less to
            # the response
            fr_gcl_cath = np.maximum(0, -fr_gcl)
            fr_gcl_anod = self.aweight * np.maximum(0, fr_gcl)
            resp_gcl = np.maximum(0, fr_gcl_cath + fr_gcl_anod - ca[1, :])
        else:
            resp_gcl = np.zeros_like(ecm[1])

        resp = resp_gcl + self.lweight * resp_inl
        resp = self.stationary_nonlinearity(resp)
        resp = self.slow_response(resp)
        return utils.TimeSeries(self.tsample, resp)


def ret2dva(r_um):
    """Converts retinal distances (um) to visual angles (deg)

    This function converts an eccentricity measurement on the retinal
    surface(in micrometers), measured from the optic axis, into degrees
    of visual angle.
    Source: Eq. A6 in Watson(2014), J Vis 14(7): 15, 1 - 17
    """
    sign = np.sign(r_um)
    r_mm = 1e-3 * np.abs(r_um)
    r_deg = 3.556 * r_mm + 0.05993 * r_mm ** 2 - 0.007358 * r_mm ** 3
    r_deg += 3.027e-4 * r_mm ** 4
    return sign * r_deg


def dva2ret(r_deg):
    """Converts visual angles (deg) into retinal distances (um)

    This function converts a retinal distancefrom the optic axis(um)
    into degrees of visual angle.
    Source: Eq. A5 in Watson(2014), J Vis 14(7): 15, 1 - 17
    """
    sign = np.sign(r_deg)
    r_deg = np.abs(r_deg)
    r_mm = 0.268 * r_deg + 3.427e-4 * r_deg ** 2 - 8.3309e-6 * r_deg ** 3
    r_um = 1e3 * r_mm
    return sign * r_um


def jansonius2009(phi0, n_rho=801, rho_range=(4.0, 45.0), eye='RE',
                  loc_od=(15.5, 1.5), beta_sup=-1.9, beta_inf=0.5):
    """Grows a single axon bundle based on the model by Jansonius et al. (2009)

    This function generates the trajectory of a single nerve fiber bundle
    based on the mathematical model described in [1]_.

    Parameters
    ----------
    phi0: float
        Angular position of the axon at its starting point(polar
        coordinates, degrees). Must be within[-180, 180].
    n_rho: int, optional, default: 801
        Number of sampling points along the radial axis(polar coordinates).
    rho_range: (rho_min, rho_max), optional, default: (4.0, 45.0)
        Lower and upper bounds for the radial position values(polar
        coordinates).
    loc_od: (x_od, y_od), optional, default: (15.5, 1.5)
        Location of the center of the optic disc(x, y) in Cartesian
        coordinates. In a right (left) eye, we should have x > 0 (x < 0).
    beta_sup: float, optional, default: -1.9
        Scalar value for the superior retina(see Eq. 5, `\beta_s` in the
        paper).
    beta_inf: float, optional, default: 0.5
        Scalar value for the inferior retina(see Eq. 6, `\beta_i` in the
        paper.)

    Returns
    -------
    ax_pos: Nx2 array
        Returns a two - dimensional array of axonal positions, where
        ax_pos[0, :] contains the(x, y) coordinates of the axon segment closest
        to the optic disc, and aubsequent row indices move the axon away from
        the optic disc. Number of rows is at most `n_rho`, but might be smaller
        if the axon crosses the meridian.

    Notes
    -----
    The study did not include axons with phi0 in [-60, 60] deg.

    .. [1] N. M. Jansionus, J. Nevalainen, B. Selig, L.M. Zangwill, P.A.
           Sample, W. M. Budde, J. B. Jonas, W. A. Lagrèze, P. J. Airaksinen,
           R. Vonthein, L. A. Levin, J. Paetzold, and U. Schieferd, "A
           mathematical description of nerve fiber bundle trajectories and
           their variability in the human retina. Vision Research 49:2157-2163,
           2009.

    """
    if eye.upper() not in ['LE', 'RE']:
        e_s = "Unknown eye string '%s': Choose from 'LE', 'RE'." % eye
        raise ValueError(e_s)

    if eye.upper() == 'LE':
        # The Jansonius model doesn't know about left eyes: We invert the x
        # coordinate of the optic disc here, run the model, and then invert all
        # x coordinates of all axon fibers back.
        loc_od = (-loc_od[0], loc_od[1])

    if np.abs(phi0) > 180.0:
        raise ValueError('phi0 must be within [-180, 180].')
    if n_rho < 1:
        raise ValueError('Number of radial sampling points must be >= 1.')
    if np.any(np.array(rho_range) < 0):
        raise ValueError('rho cannot be negative.')
    if rho_range[0] > rho_range[1]:
        raise ValueError('Lower bound on rho cannot be larger than the '
                         ' upper bound.')
    is_superior = phi0 > 0
    rho = np.linspace(*rho_range, num=n_rho)

    if is_superior:
        # Axon is in superior retina, compute `b` (real number) from Eq. 5:
        b = np.exp(beta_sup + 3.9 * np.tanh(-(phi0 - 121.0) / 14.0))
        # Equation 3, `c` a positive real number:
        c = 1.9 + 1.4 * np.tanh((phi0 - 121.0) / 14.0)
    else:
        # Axon is in inferior retina: compute `b` (real number) from Eq. 6:
        b = -np.exp(beta_inf + 1.5 * np.tanh(-(-phi0 - 90.0) / 25.0))
        # Equation 4, `c` a positive real number:
        c = 1.0 + 0.5 * np.tanh((-phi0 - 90.0) / 25.0)

    # Spiral as a function of `rho`:
    phi = phi0 + b * (rho - rho.min()) ** c

    # Convert to Cartesian coordinates
    xprime = rho * np.cos(np.deg2rad(phi))
    yprime = rho * np.sin(np.deg2rad(phi))

    # Find the array elements where the axon crosses the meridian
    if is_superior:
        # Find elements in inferior retina
        idx = np.where(yprime < 0)[0]
    else:
        # Find elements in superior retina
        idx = np.where(yprime > 0)[0]
    if idx.size:
        # Keep only up to first occurrence
        xprime = xprime[:idx[0]]
        yprime = yprime[:idx[0]]

    # Adjust coordinate system, having fovea=[0, 0] instead of `loc_od`=[0, 0]
    xmodel = xprime + loc_od[0]
    ymodel = yprime
    if loc_od[0] > 0:
        # If x-coordinate of optic disc is positive, use Appendix A
        idx = xprime > -loc_od[0]
    else:
        # Else we need to flip the sign
        idx = xprime < -loc_od[0]
    ymodel[idx] = yprime[idx] + loc_od[1] * (xmodel[idx] / loc_od[0]) ** 2

    # In a left eye, need to flip back x coordinates:
    if eye.upper() == 'LE':
        xmodel *= -1

    # Return as Nx2 array
    return np.vstack((xmodel, ymodel)).T


def find_closest_axon(pos_xy, axon_bundles):
    """Finds the closest axon to a 2D point

    This function finds the axon bundle closest to a 2D point `pos_xy` on the
    retina, and returns an axon that originates in `pos_xy` and projects to
    the optic disc.

    Parameters
    ----------
    pos_xy: (x, y)
        2D Cartesian coordinates of a location on the retina.
    axon_bundles: list of Nx2 arrays
        List of two - dimensional arrays containing the(x, y) coordinates of
        each axon bundle. The first row of each axon bundle is assumed to
        be closest to the optic disc, and subsequent row indices move the axon
        away from the optic disc.

    Returns
    -------
    axon: Nx2 array
        A single axon, where axon[0, :] contains the (x, y) coordinates of the
        location closest to `pos_xy`, and all subsequent rows move the axon
        closer to the optic disc.

    Notes
    -----
    The order of axonal segments in the output argument `axon` is reversed
    with respect to the axonal segments in the input argument `axon_bundles`.
    """
    xneuron, yneuron = pos_xy
    # Find the nearest axon to this pixel
    dist2 = [min((ax[:, 0] - xneuron) ** 2 + (ax[:, 1] - yneuron) ** 2)
             for ax in axon_bundles]
    axon_id = np.argmin(dist2)

    # Find the position on the axon
    ax = axon_bundles[axon_id]
    dist2 = (ax[:, 0] - xneuron) ** 2 + (ax[:, 1] - yneuron) ** 2
    pos_id = np.argmin(dist2)

    # Add all positions: from `pos_id` to the optic disc
    return axon_bundles[axon_id][pos_id:0:-1, :]


def axon_dist_from_soma(axon, xg, yg, tree=None):
    """Calculates the distance to soma for every axon segment

    For every segment of an axon, this function calculates the distance to the
    soma. The 2D coordinates of the axon are snapped to the grid using a
    nearest-neighbor tree structure.

    Parameters
    ----------
    axon: Nx2 array
        A single axon, where axon[0, :] contains the (x, y) coordinates of the
        location closest to the soma, and all subsequent rows move the axon
        away from the soma towards the optic disc.
    xg, yg: array
        meshgrid of pixel locations in units of visual angle sp
    tree : spat.cKDTree class instance, optional, default: train on `xg`, `yg`
        A kd-tree trained on `xg`, `yg` for quick nearest-neighbor lookup.

    Returns
    -------
    idx_cs : list
        Axon segment locations snapped to the grid, returned as a list of
        indices into the flat `xg`, `yg` meshgrid.
    dist : list
        Axon segment distances to the soma
    """
    if tree is None:
        # Build a nearest-neighbor tree for the coordinate system
        tree = spat.cKDTree(np.column_stack((xg.ravel(), yg.ravel())))

    # Consider only pixels within the grid, otherwise snap to grid might
    # yield unexpected results
    idx_valid = (axon[:, 0] >= xg.min()) * (axon[:, 0] <= xg.max())
    idx_valid *= (axon[:, 1] >= yg.min()) * (axon[:, 1] <= yg.max())

    # For these, find the xg, yg coordinates
    _, idx_cs = tree.query(axon[idx_valid, :])
    if len(idx_cs) == 0:
        return np.array([0]), np.array([np.inf])

    # Drop duplicates
    _, idx_cs_unique = np.unique(idx_cs, return_index=True)
    idx_cs = idx_cs[np.sort(idx_cs_unique)]
    if len(idx_cs) == 0:
        return np.array([0]), np.array([np.inf])

    # Find the location of the soma, based on the first axon segment
    _, idx_neuron = tree.query(axon[0, :])

    # Calculate the distance to soma.
    # For distance calculation, add a pixel at the location of the soma:
    idx_dist = np.insert(idx_cs, 0, idx_neuron, axis=0)

    # For every axon segment, calculate distance from soma by summing up the
    # individual distances between neighboring axon segments ("walking along
    # the axon"):
    xdiff = np.diff(xg.ravel()[idx_dist])
    ydiff = np.diff(yg.ravel()[idx_dist])
    dist = np.sqrt(np.cumsum(xdiff ** 2 + ydiff ** 2))

    return idx_cs, dist


def axon_contribution(axon_dist, current_spread, sensitivity_rule='decay',
                      contribution_rule='max', min_contribution=0.01,
                      decay_const=2.0, powermean_exp=None):
    """Determines the contribution of a single axon to the current map

    This function determines the contribution of a single axon to the current
    map based on a sensitivity rule (i.e., how the activation threshold of the
    axon differs as a function of distance from the soma), and an contribution
    rule (i.e., how the different activation thresholds along the axon are
    combined to determine the axon contribution).

    Parameters
    ----------
    axon_dist : tuple (indices, distances)
        A tuple containing a list of coordinates (indices into the retinal
        coordinates mesh grid) and distances for each axon segment.
    current_spread : 2D array
        A 2D current spread map that must have the same dimensions as the
        `xg`, `yg` meshgrid.
    sensitivity_rule : {'decay', 'Jeng2011'}, optional, default: 'decay'
        This rule specifies how the activation of the axon differs as a
        function of distance from the soma. The following options are
        available:
        - 'decay':
            Axon sensitivity decays exponentially with distance. Specify
            `decay_const` to change the steepness of the fall-off with
            distance.
        - 'Jeng2011':
            Axon sensitivity peaks near the sodium band (50um from the soma),
            then plateaus on the distal axon at roughly half of the peak
            sensitivity. See Figure 2 in Jeng, Tang, Molnar, Desai, and Fried
            (2011). The sodium channel band shapes the response to electric
            stimulation in retinal ganglion cells. J Neural Eng 8 (036022).
    contribution_rule : {'max', 'sum', 'power-mean'}, optional, default: 'max'
        This rule specifies how the activation thresholds across all axon
        segments are combined to determine the contribution of the axon to the
        current spread. The following options are available:
        - 'max':
            The axon's contribution to the current spread is equal to the max.
            sensitivity across all axon segments.
        - 'sum':
            The axon's contribution to the current spread is equal to the sum
            sensitivity across all axon segments.
        - 'mean':
            The axon's contribution to the current spread is equal to the mean
            sensitivity across all axon segments. Specify `powermean_exp` to
            change the exponent of the generalized (power) mean, calculated as
            np.mean(x ** powermean_exp) ** (1.0 / powermean_exp).
    min_contribution : float, optional, default: 0.01
        Current contributions below this value will not be counted.
    decay_const : float, optional, default: 2.0
        When `sensitivity_rule` is set to 'decay', specifies the decay constant
        of the exponential fall-off.
    powermean_exp : float, optional, default: None
        When `sensitivity_rule` is set to 'mean', specifies the exponent of the
        generalized (power) mean function. The power mean is calculated as
        np.mean(x ** powermean_exp) ** (1.0 / powermean_exp).
    """
    if contribution_rule == 'mean':
        if powermean_exp is None:
            raise ValueError("`powermean_exp` cannot be None when contribution "
                             "rule is set to 'mean'.")
        if powermean_exp <= 0.0:
            raise ValueError('`powermean_exp` must be positive.')
    else:
        if powermean_exp is not None:
            raise ValueError(("Contribution rule must be set to 'mean' in "
                              "order to change `powermean_exp` (currently "
                              "set to %s)." % contribution_rule))
    if decay_const <= 0.0:
        raise ValueError('`decay_const` must be positive.')

    # Unpack list of indices and distances for each axon segment
    idx_cs, dist = axon_dist
    idx_soma = idx_cs[0]

    # The sensitivity rule specifies how the activation thresholds differs
    # along the axon:
    if sensitivity_rule.lower() == 'decay':
        # Exponential fall-off with distance
        sensitivity = np.exp(-dist / decay_const)
    elif sensitivity_rule.lower() == 'jeng2011':
        # Roughly the inverse of Figure 2 in Jeng et al. (2011): The peak
        # sensitivity is over the sodium band, and the sensitivity of the
        # distal axon plateaus at roughly 50% of peak
        mu_gauss = ret2dva(50.0)
        std_gauss = ret2dva(20.0)
        bell = 0.7 * np.exp(-(dist - mu_gauss) ** 2 / (2 * std_gauss ** 2))
        plateau = 0.5
        soma = np.maximum(mu_gauss - dist, 0)
        sensitivity = np.maximum(0, bell - 0.001 * dist + plateau - soma)
    else:
        raise ValueError('Unknown sensitivity rule "%s"' % sensitivity_rule)

    # Effective activation of all axon segments, given by the segment's
    # sensitivity and activating current
    activation = sensitivity * current_spread.ravel()[idx_cs]

    # The contribution rule specifies how the activation values along the axon
    # are combined to determine the contribution of the axon to the current
    # spread:
    if contribution_rule.lower() == 'max':
        contribution = activation.max()
    elif contribution_rule.lower() == 'sum':
        contribution = activation.sum()
    elif contribution_rule.lower() == 'mean':
        # Generalized (power) mean
        p = powermean_exp
        contribution = np.mean(activation ** p) ** (1.0 / p)
    else:
        raise ValueError('Unknown activation rule "%s"' % contribution_rule)

    return idx_soma, contribution


@utils.deprecated(alt_func='p2p.retina.jansonius2009',
                  deprecated_version='0.3', removed_version='0.4')
def jansonius(num_cells=500, num_samples=801, center=np.array([15, 2]),
              rot=0 * np.pi / 180, scale=1, bs=-1.9, bi=.5, r0=4,
              max_samples=45, ang_range=60):
    """Implements the model of retinal axonal pathways by generating a
    matrix of(x, y) positions.

    Assumes that the fovea is at[0, 0]

    Parameters
    ----------
    num_cells: int
        Number of axons(cells).
    num_samples: int
        Number of samples per axon(spatial resolution).
    Center: 2 item array
        The location of the optic disk in dva.

    See:

    Jansonius et al., 2009, A mathematical description of nerve fiber bundle
    trajectories and their variability in the human retina, Vision Research
    """

    # Default parameters:
    #
    # r0 = 4;             %Minumum radius (optic disc size)
    #
    # center = [15,2];    %p.center of optic disc
    #
    # rot = 0*pi/180;    %Angle of rotation (clockwise)
    # scale = 1;             %Scale factor
    #
    # bs = -1.9;          %superior 'b' parameter constant
    # bi = .5;            %inferior 'c' parameter constant
    # ang_range = 60

    # sample space of superior/inferior retina, add them in a 1D array
    # superior is where ang0 > 0
    # this will be the first dimension of the meshgrid
    # inferior should go from -180 to -60? or typo in paper
    # ang0 is \phi_0
    ang0 = np.hstack([np.linspace(ang_range, 180, num_cells / 2),  # superior
                      np.linspace(-180, ang_range, num_cells / 2)])  # inferior

    # from r0=4 to max_samples=45, take num_samples=801 steps
    # this will be the second dimension of the meshgrid
    r = np.linspace(r0, max_samples, num_samples)

    # generate angle and radius matrices from vectors with meshgrid
    ang0mat, rmat = np.meshgrid(ang0, r)

    num_samples = ang0mat.shape[0]
    num_cells = ang0mat.shape[1]

    # index into axons from superior (upper) retina
    sup = ang0mat > 0

    # Set up 'b' parameter:
    b = np.zeros([num_samples, num_cells])

    # Equation 5: upper retina
    b[sup] = np.exp(
        bs + 3.9 * np.tanh(-(ang0mat[sup] - 121) / 14))

    # equation 6: lower retina
    b[~sup] = -np.exp(bi + 1.5 * np.tanh(-(-ang0mat[~sup] - 90) / 25))

    # Set up 'c' parameter:
    c = np.zeros([num_samples, num_cells])

    # equation 3 (fixed typo)
    # Paper says -(angmat-121)/14. Is the - sign the typo?
    c[sup] = 1.9 + 1.4 * np.tanh((ang0mat[sup] - 121) / 14)
    c[~sup] = 1 + .5 * np.tanh((-ang0mat[~sup] - 90) / 25)   # equation 4

    # Here's the main function: spirals as a function of r (equation 1)
    ang = ang0mat + b * (rmat - r0)**c

    # Transform to x-y coordinates
    xprime = rmat * np.cos(ang * np.pi / 180)
    yprime = rmat * np.sin(ang * np.pi / 180)

    # Find where the fibers cross the horizontal meridian
    cross = np.zeros([num_samples, num_cells])
    cross[sup] = yprime[sup] < 0
    cross[~sup] = yprime[~sup] > 0

    # Set Nans to axon paths after crossing horizontal meridian
    id = np.where(np.transpose(cross))

    curr_col = -1
    for i in range(0, len(id[0])):  # loop through axons
        if curr_col != id[0][i]:
            yprime[id[1][i]:, id[0][i]] = np.NaN
            curr_col = id[0][i]

    # Bend the image according to (the inverse) of Appendix A
    xmodel = xprime + center[0]
    ymodel = yprime
    id = xprime > -center[0]
    ymodel[id] = yprime[id] + center[1] * (xmodel[id] / center[0])**2

    #  rotate about the optic disc and scale
    x = scale * (np.cos(rot) * (xmodel - center[0]) + np.sin(rot) *
                 (ymodel - center[1])) + center[0]
    y = scale * (-np.sin(rot) * (xmodel - center[0]) + np.cos(rot) *
                 (ymodel - center[1])) + center[1]

    return x, y


@utils.deprecated(deprecated_version='0.3', removed_version='0.4')
def make_axon_map(xg, yg, jan_x, jan_y, axon_lambda=1, min_weight=0.001):
    """Retinal axon map

    Generates a mapping of how each pixel in the retina space is affected
    by stimulation of underlying ganglion cell axons.
    Parameters
    ----------
    xg, yg: array
        meshgrid of pixel locations in units of visual angle sp
    axon_lambda: float
        space constant for how effective stimulation(or 'weight') falls off
        with distance from the pixel back along the axon toward the optic disc
        (default 1 degree)
    min_weight: float
        minimum weight falloff.  default .001

    Returns
    -------
    axon_id: list
        a list, for every pixel, of the index into the pixel in xg, yg space,
        along the underlying axonal pathway.
    axon_weight: list
        a list, for every pixel, of the axon weight into the pixel in xg, yg
        space

    """
    axon_id = []
    axon_weight = []
    for idx, _ in enumerate(xg.ravel()):
        cur_xg = xg.ravel()[idx]
        cur_yg = yg.ravel()[idx]
        # find the nearest axon to this pixel
        d = (jan_x - cur_xg) ** 2 + (jan_y - cur_yg) ** 2
        cur_ax_id = np.nanargmin(d)  # index into the current axon

        # `ax_num`: which axon it is
        # `ax_pos_id0`: the point on that axon that is closest to `px`
        [ax_pos_id0, ax_num] = np.unravel_index(cur_ax_id, d.shape)

        dist = 0
        this_id = [idx]
        this_weight = [1.0]
        for ax_pos_id in range(ax_pos_id0 - 1, -1, -1):
            # Increment the distance from the starting point
            # The following calculation had a bug in them: squaring was done
            # twice
            ax = (jan_x[ax_pos_id + 1, ax_num] - jan_x[ax_pos_id, ax_num])
            ay = (jan_y[ax_pos_id + 1, ax_num] - jan_y[ax_pos_id, ax_num])
            dist += np.sqrt(ax ** 2 + ay ** 2)

            # weight falls off exponentially as distance from axon cell body
            weight = np.exp(-dist / axon_lambda)

            # find the nearest pixel to the current position along the axon
            dist_xg = np.abs(xg[0, :] - jan_x[ax_pos_id, ax_num])
            dist_yg = np.abs(yg[:, 0] - jan_y[ax_pos_id, ax_num])
            nearest_xg_id = dist_xg.argmin()
            nearest_yg_id = dist_yg.argmin()
            nearest_xg = xg[0, nearest_xg_id]
            nearest_yg = yg[nearest_yg_id, 0]

            # if the position along the axon has moved to a new pixel, and the
            # weight isn't too small...
            if weight > min_weight:
                if nearest_xg != cur_xg or nearest_yg != cur_yg:
                    # update the current pixel location
                    cur_xg = nearest_xg
                    cur_yg = nearest_yg

                    # Append the list
                    # The following calculation had a bug in it: `weight` was
                    # exponentiated twice
                    this_weight.append(weight)
                    this_id.append(np.ravel_multi_index((nearest_yg_id,
                                                         nearest_xg_id),
                                                        xg.shape))

        axon_id.append(this_id)
        axon_weight.append(this_weight)
    return axon_id, axon_weight



"""
Utility functions for pulse2percept
"""
import numpy as np
import multiprocessing
import random
import copy
import functools
import logging

from scipy import misc as spm
from scipy import interpolate as spi
from scipy import signal as sps


# Rather than trying to import these all over, try once and then remember
# by setting a flag.
try:
    import joblib
    joblib.Parallel
    has_joblib = True
except (ImportError, AttributeError):
    has_joblib = False

try:
    import dask
    import dask.multiprocessing
    dask.delayed
    has_dask = True
except (ImportError, AttributeError):
    has_dask = False

try:
    from numba import jit
    has_jit = True
except ImportError:
    has_jit = False


class deprecated(object):
    """Decorator to mark deprecated functions with warning.

    Adapted from <https://github.com/scikit-image/scikit-image/blob/master/skimage/_shared/utils.py>,
    who adapted it from <http://wiki.python.org/moin/PythonDecoratorLibrary>.

    Parameters
    ----------
    alt_func : str
        If given, tell user what function to use instead.
    behavior : {'warn', 'raise'}
        Behavior during call to deprecated function: 'warn' = warn user that
        function is deprecated; 'raise' = raise error.
    removed_version : str
        The package version in which the deprecated function will be removed.
    """

    def __init__(self, alt_func=None, behavior='warn', deprecated_version=None,
                 removed_version=None):
        self.alt_func = alt_func
        self.behavior = behavior
        self.deprecated_version = deprecated_version
        self.removed_version = removed_version

    def __call__(self, func):

        alt_msg = ""
        if self.alt_func is not None:
            alt_msg = "Use ``%s`` instead." % self.alt_func
        dep_msg = ""
        if self.deprecated_version is not None:
            dep_msg = " since version %s" % self.deprecated_version
        rmv_msg = ""
        if self.removed_version is not None:
            rmv_msg = (", and will be removed in version %s" %
                       self.removed_version)

        msg = "Function ``%s`` is deprecated" % func.__name__
        msg += dep_msg + rmv_msg + ". " + alt_msg

        @functools.wraps(func)
        def wrapped(*args, **kwargs):
            if self.behavior == 'warn':
                logging.getLogger(__name__).warn(msg)
            elif self.behavior == 'raise':
                raise RuntimeError(msg)
            return func(*args, **kwargs)

        # modify doc string to display deprecation warning
        doc = "**Deprecated function**. " + alt_msg
        doc += "\n\n    " + msg
        if wrapped.__doc__ is None:
            wrapped.__doc__ = doc
        else:
            wrapped.__doc__ = doc + "\n\n    " + wrapped.__doc__

        return wrapped


class TimeSeries(object):

    def __init__(self, tsample, data):
        """Container for time series data

        Provides a container for time series data. Every time series has a
        sampling step `tsample`, and some `data` sampled at that rate.

        Parameters
        ----------
        tsample : float
            Sampling time step (seconds).
        data : array_like
            Time series data sampled at every `tsample` seconds.
        """
        self.data = data
        self.tsample = tsample
        self.duration = self.data.shape[-1] * tsample
        self.shape = data.shape

    def __getitem__(self, y):
        return TimeSeries(self.tsample, self.data[y])

    def append(self, other):
        """Appends the data of another TimeSeries object (in time)

        This function concatenates the data of another TimeSeries object to
        the current object, along the last dimension (time). To make this work,
        all but the last dimension of the two objects must be the same.

        If the two objects have different time sampling steps, the other object
        is resampled to fit the current `tsample`.

        Parameters
        ----------
        other : p2p.utils.TimeSeries
            A TimeSeries object whose content should be appended.

        Examples
        --------
        >>> from pulse2percept import utils
        >>> pt = utils.TimeSeries(1.0, np.zeros((2, 2, 10)))
        >>> num_frames = pt.shape[-1]
        >>> pt.append(pt)
        >>> pt.shape[-1] == 2 * num_frames
        True
        """
        # Make sure type is correct
        if not isinstance(other, TimeSeries):
            raise TypeError("Other object must be of type "
                            "p2p.utils.TimeSeries.")

        # Make sure size is correct for all but the last dimension (number
        # of frames)
        if self.shape[:-1] != other.shape[:-1]:
            raise ValueError("Shape mismatch: ", self.shape[:-1], " vs. ",
                             other.shape[:-1])

        # Then resample the other to current `tsample`
        resampled = other.resample(self.tsample)

        # Then concatenate the two
        self.data = np.concatenate((self.data, resampled.data), axis=-1)
        self.duration = self.data.shape[-1] * self.tsample
        self.shape = self.data.shape

    def max(self):
        """Returns the time and value of the largest data point

        This function returns the the largest value in the TimeSeries data,
        as well as the time at which it occurred.

        Returns
        -------
        t : float
            time (s) at which max occurred
        val : float
            max value
        """
        # Find index and value of largest element
        idx = self.data.argmax()
        val = self.data.max()

        # Find frame that contains the brightest data point using `unravel`,
        # which maps the flat index `idx_px` onto the high-dimensional
        # indices (x,y,z).
        # What we want is index `z` (i.e., the frame index), given by the last
        # dimension in the return argument.
        idx_frame = np.unravel_index(idx, self.data.shape)[-1]

        # Convert index to time
        t = idx_frame * self.tsample

        return t, val

    def max_frame(self):
        """Returns the time frame that contains the largest data point

        This function returns the time frame in the TimeSeries data that
        contains the largest data point, as well as the time at which
        it occurred.

        Returns
        -------
        t : float
            time (s) at which max occurred
        frame : TimeSeries
            A TimeSeries object.
        """
        # Find index and value of largest element
        idx = self.data.argmax()

        # Find frame that contains the brightest data point using `unravel`,
        # which maps the flat index `idx_px` onto the high-dimensional
        # indices (x,y,z).
        # What we want is index `z` (i.e., the frame index), given by the last
        # dimension in the return argument.
        idx_frame = np.unravel_index(idx, self.data.shape)[-1]
        t = idx_frame * self.tsample
        frame = self.data[..., idx_frame]

        return t, TimeSeries(self.tsample, copy.deepcopy(frame))

    def resample(self, tsample_new):
        """Returns data sampled according to new time step

        This function returns a TimeSeries object whose data points were
        resampled according to a new time step `tsample_new`. New values
        are found using linear interpolation.

        Parameters
        ----------
        tsample_new : float
            New sampling time step (s)

        Returns
        -------
        TimeSeries object
        """
        if tsample_new is None or tsample_new == self.tsample:
            return TimeSeries(self.tsample, self.data)

        # Try to avoid rounding errors in arr size by making sure `t_old` is
        # too long at first, then cutting it to the right size
        y_old = self.data
        t_old = np.arange(0, self.duration + self.tsample, self.tsample)
        t_old = t_old[:y_old.shape[-1]]
        f = spi.interp1d(t_old, y_old, axis=-1, fill_value='extrapolate')

        t_new = np.arange(0, self.duration, tsample_new)
        y_new = f(t_new)

        return TimeSeries(tsample_new, y_new)


def center_vector(vec, newlen):
    """
    Returns the center `newlen` portion of a vector.

    Adapted from scipy.signal.signaltools._centered:
    github.com/scipy/scipy/blob/v0.18.0/scipy/signal/signaltools.py#L236-L243

    """
    currlen = vec.shape[-1]
    startind = (currlen - newlen) // 2
    endind = startind + newlen
    return vec[startind:endind]


def _sparseconv(data, kernel, mode):
    """Returns the discrete, linear convolution of two 1D sequences

    This function returns the discrete, linear convolution of two
    one-dimensional sequences, where the length of the output is determined
    by `mode`.
    Can run faster than ``np.convolve`` if:
    (1) `data` is much longer than `kernel`
    (2) `data` is sparse (has lots of zeros)
    """
    kernel_len = kernel.size
    data_len = data.size
    out = np.zeros(data_len + kernel_len - 1)

    pos = np.where(data.ravel() != 0)[0]
    # Add shifted and scaled copies of `kernel` only where `data` is nonzero
    for p in pos:
        out[p:p + kernel_len] = (out[p:p + kernel_len] +
                                 kernel.ravel() * data.ravel()[p])

    if mode.lower() == 'full':
        return out
    elif mode.lower() == 'valid':
        return center_vector(out, data_len - kernel_len + 1)
    elif mode.lower() == 'same':
        return center_vector(out, data_len)
    else:
        raise ValueError("Acceptable mode flags are 'valid',"
                         " 'same', or 'full'.")


def sparseconv(data, kernel, mode, use_jit=True):
    """Convolves data with a kernel using sparse convolution

    This function convolves data with a kernel, relying either on the
    fast Fourier transform (FFT) or a sparse convolution function.

    Parameters
    ----------
    data : array_like
        First input, typically the data array
    kernel : array_like
        Second input, typically the kernel
    mode : str {'full', 'valid', 'same'}, optional, default: 'full'
        A string indicating the size of the output:

        - ``full``, default
            The output is the full discrete linear convolution of the inputs.
        - ``valid``:
            The output consists only of those elements that do not rely on
            zero-padding.
        - ``same``:
            The output is the same size as `data`, centered with respect to the
            'full' output.

    method : str {'fft', 'sparse'}, optional, default: 'fft'
        A string indicating the convolution method:

        - ``fft``: default
            Use the fast Fourier transform (FFT).
        - ``sparse``:
            Use the sparse convolution.

    use_jit : bool, optional, default: True
        A flag indicating whether to use Numba's just-in-time compilation
        option.
    """
    func_sparseconv = _sparseconv
    if use_jit:
        if not has_jit:
            e_s = "You do not have numba, please run sparseconv with "
            e_s += "`use_jit`=False."
            raise ImportError(e_s)
        func_sparseconv = jit(_sparseconv)
    return func_sparseconv(data, kernel, mode)


def conv(data, kernel, mode='full', method='fft', use_jit=True):
    """Convoles data with a kernel using either FFT or sparse convolution

    This function convolves data with a kernel, relying either on the
    fast Fourier transform (FFT) or a sparse convolution function.

    Parameters
    ----------
    data : array_like
        First input, typically the data array
    kernel : array_like
        Second input, typically the kernel
    mode : str {'full', 'valid', 'same'}, optional, default: 'full'
        A string indicating the size of the output:

        - ``full``:
            The output is the full discrete linear convolution of the inputs.
        - ``valid``:
            The output consists only of those elements that do not rely on
            zero-padding.
        - ``same``:
            The output is the same size as `data`, centered with respect to the
            'full' output.

    method : str {'fft', 'sparse'}, optional, default: 'fft'
        A string indicating the convolution method:

        - ``fft``:
            Use the fast Fourier transform (FFT).
        - ``sparse``:
            Use the sparse convolution.

    use_jit : bool, optional, default: True
        A flag indicating whether to use Numba's just-in-time compilation
        option (only relevant for `method`=='sparse').
    """
    if method.lower() == 'fft':
        # Use FFT: faster on non-sparse data
        conved = sps.fftconvolve(data, kernel, mode)
    elif method.lower() == 'sparse':
        # Use sparseconv: faster on sparse data
        conved = sparseconv(data, kernel, mode, use_jit)
    else:
        raise ValueError("Acceptable methods are: 'fft', 'sparse'.")
    return conved


def parfor(func, in_list, out_shape=None, n_jobs=-1, engine='joblib',
           scheduler='threading', func_args=[], func_kwargs={}):
    """
    Parallel for loop for numpy arrays

    Parameters
    ----------
    func : callable
        The function to apply to each item in the array. Must have the form:
        func(arr, idx, *args, *kwargs) where arr is an ndarray and idx is an
        index into that array (a tuple). The Return of `func` needs to be one
        item (e.g. float, int) per input item.
    in_list : list
       All legitimate inputs to the function to operate over.
    out_shape : int or tuple of ints, optional
        If set, output will be reshaped accordingly. The new shape should be
        compatible with the original shape. If an integer, then the result will
        be a 1-D array of that length. One shape dimension can be -1. In this
        case, the value is inferred from the length of the array and remaining
        dimensions.
    n_jobs : integer, optional, default: 1
        The number of jobs to perform in parallel. -1 to use all cpus
    engine : str, optional, default: 'joblib'
        {'dask', 'joblib', 'serial'}
        The last one is useful for debugging -- runs the code without any
        parallelization.
    scheduler : str, optional, default: 'threading'
        Which scheduler to use (irrelevant for 'serial' engine):
        - 'threading': a scheduler backed by a thread pool
        - 'multiprocessing': a scheduler backed by a process pool
    func_args : list, optional
        Positional arguments to `func`
    func_kwargs : list, optional
        Keyword arguments to `func`

    Returns
    -------
    ndarray of identical shape to `arr`

    Notes
    -----
    Equivalent to pyAFQ version (blob e20eaa0 from June 3, 2016):
    https://github.com/arokem/pyAFQ/blob/master/AFQ/utils/parallel.py

    Examples
    --------
    """
    if n_jobs == -1:
        n_jobs = multiprocessing.cpu_count()
        n_jobs = n_jobs - 1

    if engine.lower() == 'joblib':
        if not has_joblib:
            err = "You do not have `joblib` installed. Consider setting"
            err += "`engine` to 'serial' or 'dask'."
            raise ImportError(err)

        p = joblib.Parallel(n_jobs=n_jobs, backend=scheduler)
        d = joblib.delayed(func)
        d_l = []
        for in_element in in_list:
            d_l.append(d(in_element, *func_args, **func_kwargs))
        results = p(d_l)
    elif engine.lower() == 'dask':
        if not has_dask:
            err = "You do not have `dask` installed. Consider setting"
            err += "`engine` to 'serial' or 'joblib'."
            raise ImportError(err)

        def partial(func, *args, **keywords):
            def newfunc(in_arg):
                return func(in_arg, *args, **keywords)
            return newfunc
        p = partial(func, *func_args, **func_kwargs)
        d = [dask.delayed(p)(i) for i in in_list]
        if scheduler == 'multiprocessing':
            results = dask.compute(*d, get=dask.multiprocessing.get,
                                   workers=n_jobs)
        elif scheduler == 'threading':
            results = dask.compute(*d, get=dask.threaded.get, workers=n_jobs)
        else:
            raise ValueError("Acceptable values for `scheduler` are: "
                             "'threading', 'multiprocessing'")
    elif engine.lower() == 'serial':
        results = []
        for in_element in in_list:
            results.append(func(in_element, *func_args, **func_kwargs))
    else:
        raise ValueError("Acceptable values for `engine` are: 'serial', "
                         "'joblib', or 'dask'.")

    if out_shape is not None:
        return np.array(results).reshape(out_shape)
    else:
        return results


def gamma(n, tau, tsample, tol=0.01):
    """Returns the impulse response of `n` cascaded leaky integrators

    This function calculates the impulse response of `n` cascaded
    leaky integrators with constant of proportionality 1/`tau`:
    y = (t/theta).^(n-1).*exp(-t/theta)/(theta*factorial(n-1))

    Parameters
    ----------
    n : int
        Number of cascaded leaky integrators
    tau : float
        Decay constant of leaky integration (seconds).
        Equivalent to the inverse of the constant of proportionality.
    tsample : float
        Sampling time step (seconds).
    tol : float
        Cut the kernel to size by ignoring function values smaller
        than a fraction `tol` of the peak value.
    """
    n = int(n)
    tau = float(tau)
    tsample = float(tsample)
    if n <= 0 or tau <= 0 or tsample <= 0:
        raise ValueError("`n`, `tau`, and `tsample` must be nonnegative.")
    if tau <= tsample:
        raise ValueError("`tau` cannot be smaller than `tsample`.")

    # Allocate a time vector that is long enough for sure.
    # Trim vector later on.
    t = np.arange(0, 5 * n * tau, tsample)

    # Calculate gamma
    y = (t / tau) ** (n - 1) * np.exp(-t / tau)
    y /= (tau * spm.factorial(n - 1))

    # Normalize to unit area
    y /= np.trapz(np.abs(y), dx=tsample)

    # Cut off tail where values are smaller than `tol`.
    # Make sure to start search on the right-hand side of the peak.
    peak = y.argmax()
    small_vals = np.where(y[peak:] < tol * y.max())[0]
    if small_vals.size:
        t = t[:small_vals[0] + peak]
        y = y[:small_vals[0] + peak]

    return t, y


def traverse_randomly(seq):
    """Traverses a list in random order

    Parameters
    ----------
    seq : list
        An iterable list of elements.
    Returns
    -------
    A list iterator.

    Examples
    --------
    Shuffle a list:

    >>> list_ordered = [0, 1, 2, 3]
    >>> list_shuffled = [l for l in traverse_randomly(list_ordered)]

    Traverse a list in random order:

    >>> list_ordered = [0, 1, 2, 3]
    >>> for idx, val in traverse_randomly(enumerate(list_ordered)):
    ...     pass

    Notes
    -----
    From: http://stackoverflow.com/a/9253366
    Note that even for rather small `len(x)`, the total number of permutations
    of `seq` can quickly grow larger than the period of most random number
    generators. For example, a sequence of length 2080 is the largest that can
    fit within the period of the Mersenne Twister random number generator.
    """
    # Make sure we are dealing with a list
    shuffled = list(seq)

    # Import `random` at module level (for performance)
    random.shuffle(shuffled)

    return iter(shuffled)



import numpy as np
import numpy.testing as npt
import pytest

from .. import implants


def test_Electrode():
    num_pts = 10
    r = np.linspace(1, 1000, num_pts)
    x = np.linspace(-1000, 1000, num_pts)
    y = np.linspace(-2000, 2000, num_pts)
    h = np.linspace(0, 1000, num_pts)
    t = ['subretinal', 'epiretinal'] * (num_pts // 2)
    n = ["some name"] * num_pts

    for rr, xx, yy, hh, tt, nn in zip(r, x, y, h, t, n):
        e = implants.Electrode(tt, rr, xx, yy, hh, nn)
        npt.assert_equal(e.radius, rr)
        npt.assert_equal(e.x_center, xx)
        npt.assert_equal(e.y_center, yy)
        # npt.assert_equal(e.h, hh)
        npt.assert_equal(e.etype, tt)
        npt.assert_equal(e.name, nn)
        if tt.lower() == 'epiretinal':
            # `height` property should return `h_ofl`
            npt.assert_equal(e.height, e.h_ofl)

            # `h_ofl` should be the same as the user-specified height
            npt.assert_equal(e.height, hh)

            # `h_inl` should be further away from the array
            npt.assert_equal(e.h_inl > hh, True)
        else:
            # `height` property should return `h_inl`
            npt.assert_equal(e.height, e.h_inl)

            # Subretinal arrays have layer thicknesses added to `hh`.
            npt.assert_equal(e.height > hh, True)

    # Invalid type
    with pytest.raises(ValueError):
        implants.Electrode('suprachoroidal', 10, 0, 0, 0)

    # Invalid layer
    e = implants.Electrode('epiretinal', 10, 0, 0, 0)
    with pytest.raises(ValueError):
        e.current_spread(0, 0, 'RGC')

    # Invalid type lookup
    e.etype = 'suprachoroidal'
    with pytest.raises(ValueError):
        e.height


def test_ElectrodeArray():
    implant = implants.ElectrodeArray('subretinal', 10, 0, 0)
    npt.assert_equal(implant.num_electrodes, 1)

    # Make sure ElectrodeArray can accept ints, floats, lists, np.arrays
    arrays = [None] * 4
    arrays[0] = implants.ElectrodeArray('epiretinal', [0], [1], [2],
                                        hs=[3])
    arrays[1] = implants.ElectrodeArray('epiretinal', 0, 1, 2, hs=3)
    arrays[2] = implants.ElectrodeArray('epiretinal', .0, [1], 2.0,
                                        hs=[3])
    arrays[3] = implants.ElectrodeArray('epiretinal', np.array([0]), [1],
                                        [2], hs=[[3]])

    for arr in arrays:
        npt.assert_equal(arr.num_electrodes, 1)
        npt.assert_equal(arr.electrodes[0].radius, 0)
        npt.assert_equal(arr.electrodes[0].x_center, 1)
        npt.assert_equal(arr.electrodes[0].y_center, 2)
        npt.assert_equal(arr.electrodes[0].h_ofl, 3)
        npt.assert_equal(arr.electrodes[0].etype, 'epiretinal')

    # Make sure electrodes can be addressed by index
    vals = range(5)
    implant = implants.ElectrodeArray('subretinal', vals, vals, vals,
                                      hs=vals)
    npt.assert_equal(implant.num_electrodes, len(vals))
    for v in vals:
        el = implant[v]
        npt.assert_equal(el.radius, v)
        npt.assert_equal(el.x_center, v)
        npt.assert_equal(el.y_center, v)
        npt.assert_equal(el.h_inl, v + 23.0 / 2.0)
        npt.assert_equal(el.h_ofl, v + 83.0)

    # Test left/right eye
    for validstr in ['left', 'LEFT', 'l', 'LE']:
        implant = implants.ElectrodeArray('epiretinal', 10, 0, 0,
                                          eye=validstr)
        npt.assert_equal(implant.eye, 'LE')
    for validstr in ['right', 'Right', 'r', 'RE']:
        implant = implants.ElectrodeArray('epiretinal', 10, 0, 0,
                                          eye=validstr)
        npt.assert_equal(implant.eye, 'RE')
    for invalidstr in ['both', 'lefteye', 'invalid']:
        with pytest.raises(ValueError):
            implant = implants.ElectrodeArray('epiretinal', 10, 0, 0,
                                              eye=invalidstr)


def test_ElectrodeArray_add_electrode():
    implant = implants.ElectrodeArray('epiretinal', 10, 0, 0)
    with pytest.raises(TypeError):
        implant.add_electrode(implant)

    with pytest.raises(ValueError):
        implant.add_electrode(implants.Electrode('subretinal', 10, 0, 0))

    # Make sure electrode count is correct
    for j in range(5):
        implant.add_electrode(implants.Electrode('epiretinal', 10, 10, 10))
        npt.assert_equal(implant.num_electrodes, j + 2)


def test_ElectrodeArray_add_electrodes():
    for j in range(5):
        implant = implants.ElectrodeArray('epiretinal', 10, 0, 0)
        implant.add_electrodes(range(1, j + 1), range(j), range(j))
        npt.assert_equal(implant.num_electrodes, j + 1)

    # However, all input arguments must have the same number of elements
    with pytest.raises(AssertionError):
        implant.add_electrodes([0], [1, 2], [3, 4, 5], [6])


def test_ArgusI():
    # Create an ArgusI and make sure location is correct
    for htype in ['float', 'list']:
        for x in [0, -100, 200]:
            for y in [0, -200, 400]:
                for r in [0, -30, 45, 60, -90]:
                    # Height `h` can either be a float or a list
                    if htype == 'float':
                        h = 100
                    else:
                        h = np.ones(16) * 20

                    # Convert rotation angle to rad
                    rot = r * np.pi / 180
                    argus = implants.ArgusI(x, y, h=h, rot=rot)

                    # Coordinates of first electrode
                    xy = np.array([-1200, -1200]).T

                    # Rotate
                    R = np.array([np.cos(rot), -np.sin(rot),
                                  np.sin(rot), np.cos(rot)]).reshape((2, 2))
                    xy = np.matmul(R, xy)

                    # Then off-set: Make sure first electrode is placed
                    # correctly
                    npt.assert_almost_equal(argus['A1'].x_center,
                                            xy[0] + x)
                    npt.assert_almost_equal(argus['A1'].y_center,
                                            xy[1] + y)

                    # Make sure array center is still (x,y)
                    y_center = argus['D1'].y_center + \
                        (argus['A4'].y_center - argus['D1'].y_center) / 2
                    npt.assert_almost_equal(y_center, y)
                    x_center = argus['A1'].x_center + \
                        (argus['D4'].x_center - argus['A1'].x_center) / 2
                    npt.assert_almost_equal(x_center, x)

                    # Check radii of electrodes
                    for e in ['A1', 'A3', 'B2', 'C1', 'D4']:
                        npt.assert_almost_equal(argus[e].radius, 130)
                    for e in ['A2', 'A4', 'B1', 'C2', 'D3']:
                        npt.assert_almost_equal(argus[e].radius, 260)

                    # Check location of the tack
                    tack = np.matmul(R, [-2000, 0])
                    tack = tuple(tack + [x_center, y_center])

    # `h` must have the right dimensions
    with pytest.raises(ValueError):
        implants.ArgusI(-100, 10, h=np.zeros(5))
    with pytest.raises(ValueError):
        implants.ArgusI(-100, 10, h=[1, 2, 3])

    for use_legacy_names in [False, True]:
        # Indexing must work for both integers and electrode names
        argus = implants.ArgusI(use_legacy_names=use_legacy_names)
        for idx, electrode in enumerate(argus):
            name = electrode.name
            npt.assert_equal(electrode, argus[idx])
            npt.assert_equal(electrode, argus[name])
        npt.assert_equal(argus[16], None)
        npt.assert_equal(argus["unlikely name for an electrode"], None)

        if use_legacy_names:
            name_idx1 = 'L2'
            name_idx4 = 'L5'
        else:
            name_idx1 = 'B1'
            name_idx4 = 'A2'

        # Indexing must have the right order
        npt.assert_equal(argus.get_index(name_idx1), 1)
        npt.assert_equal(argus[name_idx1], argus[1])
        npt.assert_equal(argus.get_index(name_idx4), 4)
        npt.assert_equal(argus[name_idx4], argus[4])

    # Right-eye implant:
    xc, yc = 500, -500
    argus_re = implants.ArgusI(eye='RE', x_center=xc, y_center=yc)
    npt.assert_equal(argus_re['D1'].x_center > argus_re['A1'].x_center, True)
    npt.assert_almost_equal(argus_re['D1'].y_center, argus_re['A1'].y_center)
    npt.assert_equal(argus_re.tack[0] < argus_re['D1'].x_center, True)
    npt.assert_almost_equal(argus_re.tack[1], yc)

    # Left-eye implant:
    argus_le = implants.ArgusI(eye='LE', x_center=xc, y_center=yc)
    npt.assert_equal(argus_le['A1'].x_center > argus_le['D1'].x_center, True)
    npt.assert_almost_equal(argus_le['D1'].y_center, argus_le['A1'].y_center)
    npt.assert_equal(argus_le.tack[0] > argus_le['A1'].x_center, True)
    npt.assert_almost_equal(argus_le.tack[1], yc)

    # In both left and right eyes, rotation with positive angle should be
    # counter-clock-wise (CCW): for (x>0,y>0), decreasing x and increasing y
    for eye, el in zip(['LE', 'RE'], ['A1', 'D4']):
        before = implants.ArgusI(eye=eye)
        after = implants.ArgusI(eye=eye, rot=np.deg2rad(10))
        npt.assert_equal(after[el].x_center < before[el].x_center, True)
        npt.assert_equal(after[el].y_center > before[el].y_center, True)

    argus = implants.ArgusI()
    # Old to new
    npt.assert_equal(argus.get_new_name('M1'), 'D4')
    npt.assert_equal(argus.get_new_name('M6'), 'C3')
    # New to old
    npt.assert_equal(argus.get_old_name('B2'), 'L1')
    npt.assert_equal(argus.get_old_name('A1'), 'L6')


def test_ArgusII():
    # Create an ArgusII and make sure location is correct
    for htype in ['float', 'list']:
        for x in [0, -100, 200]:
            for y in [0, -200, 400]:
                for r in [0, -30, 45, 60, -90]:
                    # Height `h` can either be a float or a list
                    if htype == 'float':
                        h = 100
                    else:
                        h = np.ones(60) * 20

                    # Convert rotation angle to rad
                    rot = np.deg2rad(r)
                    argus = implants.ArgusII(x, y, h=h, rot=rot)

                    # Coordinates of first electrode
                    xy = np.array([-2362.5, -1312.5]).T

                    # Rotate
                    R = np.array([np.cos(rot), -np.sin(rot),
                                  np.sin(rot), np.cos(rot)]).reshape((2, 2))
                    xy = np.matmul(R, xy)

                    # Then off-set: Make sure first electrode is placed
                    # correctly
                    npt.assert_almost_equal(argus['A1'].x_center,
                                            xy[0] + x)
                    npt.assert_almost_equal(argus['A1'].y_center,
                                            xy[1] + y)

                    # Make sure array center is still (x,y)
                    y_center = argus['F1'].y_center + \
                        (argus['A10'].y_center - argus['F1'].y_center) / 2
                    npt.assert_almost_equal(y_center, y)
                    x_center = argus['A1'].x_center + \
                        (argus['F10'].x_center - argus['A1'].x_center) / 2
                    npt.assert_almost_equal(x_center, x)

                    # Make sure radius is correct
                    for e in ['A1', 'B3', 'C5', 'D7', 'E9', 'F10']:
                        npt.assert_almost_equal(argus[e].radius, 100)

    # `h` must have the right dimensions
    with pytest.raises(ValueError):
        implants.ArgusII(-100, 10, h=np.zeros(5))
    with pytest.raises(ValueError):
        implants.ArgusII(-100, 100, h=[1, 2, 3])

    # Indexing must work for both integers and electrode names
    argus = implants.ArgusII()
    for idx, electrode in enumerate(argus):
        name = electrode.name
        npt.assert_equal(electrode, argus[idx])
        npt.assert_equal(electrode, argus[name])
    npt.assert_equal(argus[60], None)
    npt.assert_equal(argus["unlikely name for an electrode"], None)

    # Indexing must have the right order
    npt.assert_equal(argus.get_index('A2'), 1)
    npt.assert_equal(argus['A2'], argus[1])
    npt.assert_equal(argus.get_index('B1'), 10)
    npt.assert_equal(argus['B1'], argus[10])

    # Right-eye implant:
    xc, yc = 500, -500
    argus_re = implants.ArgusII(eye='RE', x_center=xc, y_center=yc)
    npt.assert_equal(argus_re['A10'].x_center > argus_re['A1'].x_center, True)
    npt.assert_almost_equal(argus_re['A10'].y_center, argus_re['A1'].y_center)
    npt.assert_equal(argus_re.tack[0] < argus_re['A1'].x_center, True)
    npt.assert_almost_equal(argus_re.tack[1], yc)

    # Left-eye implant:
    argus_le = implants.ArgusII(eye='LE', x_center=xc, y_center=yc)
    npt.assert_equal(argus_le['A1'].x_center > argus_le['A10'].x_center, True)
    npt.assert_almost_equal(argus_le['A10'].y_center, argus_le['A1'].y_center)
    npt.assert_equal(argus_le.tack[0] > argus_le['A10'].x_center, True)
    npt.assert_almost_equal(argus_le.tack[1], yc)

    # In both left and right eyes, rotation with positive angle should be
    # counter-clock-wise (CCW): for (x>0,y>0), decreasing x and increasing y
    for eye, el in zip(['LE', 'RE'], ['F1', 'F10']):
        before = implants.ArgusII(eye=eye)
        after = implants.ArgusII(eye=eye, rot=np.deg2rad(10))
        npt.assert_equal(after[el].x_center < before[el].x_center, True)
        npt.assert_equal(after[el].y_center > before[el].y_center, True)


def test_Electrode_receptive_field():
    electrode = implants.Electrode('epiretinal', 100, 0, 0, 0)

    with pytest.raises(ValueError):
        electrode.receptive_field(0, 0, rftype='invalid')



from __future__ import absolute_import, division, print_function
from os.path import join as pjoin

# Format expected by setup.py and doc/source/conf.py: string of form "X.Y.Z"
_version_major = 0
_version_minor = 4
_version_micro = 2  # use '' for first of series, number for 1 and above
_version_extra = ''
# _version_extra = ''  # Uncomment this for full releases

# Construct full version string from these.
_ver = [_version_major, _version_minor]
if _version_micro:
    _ver.append(_version_micro)
if _version_extra:
    _ver.append(_version_extra)

__version__ = '.'.join(map(str, _ver))

CLASSIFIERS = ["Development Status :: 4 - Beta",
               "Environment :: Console",
               "Intended Audience :: Science/Research",
               "License :: OSI Approved :: BSD License",
               "Operating System :: OS Independent",
               "Programming Language :: Python",
               "Topic :: Scientific/Engineering"]

# Description should be a one-liner:
description = "pulse2percept: A Python-based simulation framework for bionic vision"
# Long description will go up on the pypi page
long_description = """

pulse2percept: A Python-based simulation framework for bionic vision
--------------------------------------------------------------------

By 2020 roughly 200 million people will suffer from retinal diseases such as
macular degeneration or retinitis pigmentosa. Consequently, a variety of
retinal sight restoration procedures are being developed to target these
diseases. Electronic prostheses (currently being implanted in patients)
directly stimulate remaining retinal cells using electrical current, analogous
to a cochlear implant. Optogenetic prostheses (soon to be implanted in human)
use optogenetic proteins to make remaining retinal cells responsive to light,
then use light diodes (natural illumination is inadequate) implanted in the
eye to stimulate these light sensitive cells.

However, these devices do not restore anything resembling natural vision:
Interactions between the electronics and the underlying neurophysiology result
in significant distortions of the perceptual experience.

We have developed a computer model that has the goal of predicting the
perceptual experience of retinal prosthesis patients. The model was developed
using a variety of patient data describing the brightness and shape of
phosphenes elicited by stimulating single electrodes, and validated against an
independent set of behavioral measures examining spatiotemporal interactions
across multiple electrodes.

More information can be found in
`Beyeler et al. (2017) <https://doi.org/10.25080/shinma-7f4c6e7-00c>`_
and in our
`Github repo <https://github.com/uwescience/pulse2percept>`_.

"""

NAME = "pulse2percept"
MAINTAINER = "Michael Beyeler, Ariel Rokem"
MAINTAINER_EMAIL = "mbeyeler@uw.edu, arokem@gmail.com"
DESCRIPTION = description
LONG_DESCRIPTION = long_description
URL = "http://github.com/uwescience/pulse2percept"
DOWNLOAD_URL = ""
LICENSE = "BSD"
AUTHOR = "Michael Beyeler, Ariel Rokem"
AUTHOR_EMAIL = "mbeyeler@uw.edu, arokem@gmail.com"
PLATFORMS = "OS Independent"
MAJOR = _version_major
MINOR = _version_minor
MICRO = _version_micro
VERSION = __version__
PACKAGE_DATA = {}
REQUIRES = ["numpy", "scipy", "joblib", "scikit_image", "sk_video", "cython"]



import numpy as np
import numpy.testing as npt
import pytest

from .. import api as p2p
from .. import implants
from .. import stimuli
from .. import utils


def test_Simulation___init__():
    implant = implants.Electrode("epiretinal", 10, 0, 0, 0)
    with pytest.raises(TypeError):
        p2p.Simulation(implant)


def test_Simulation_pulse2percept():
    implant = implants.ElectrodeArray("epiretinal", 10, 0, 0, 0)
    sim = p2p.Simulation(implant, engine='serial')
    sim.set_optic_fiber_layer(x_range=[0, 0], y_range=[0, 0])
    pt = stimuli.BiphasicPulse('cathodicfirst', 0.45 / 1000, 0.005 / 1000)
    sim.pulse2percept(pt)
    sim.pulse2percept(pt, layers=['GCL'])
    sim.pulse2percept(pt, layers=['INL'])

    # PulseTrain must have the same tsample as (implicitly set up) GCL
    pt = stimuli.BiphasicPulse("cathodicfirst", 0.1, 0.001)
    with pytest.raises(ValueError):
        sim.pulse2percept(pt)

    pt = stimuli.BiphasicPulse("cathodicfirst", 0.1, 0.005 / 1000)
    with pytest.raises(ValueError):
        sim.pulse2percept(pt, layers=['GCL', 'invalid'])


def test_Simulation_set_optic_fiber_layer():
    sim = p2p.Simulation(implants.ArgusI(), engine='serial')

    # Invalid grid ranges
    with pytest.raises(ValueError):
        sim.set_optic_fiber_layer(x_range=(10, 0))
    with pytest.raises(ValueError):
        sim.set_optic_fiber_layer(x_range=(1, 2, 3))
    with pytest.raises(ValueError):
        sim.set_optic_fiber_layer(x_range='invalid')
    with pytest.raises(ValueError):
        sim.set_optic_fiber_layer(y_range=(10, 0))
    with pytest.raises(ValueError):
        sim.set_optic_fiber_layer(y_range=(1, 2, 3))
    with pytest.raises(ValueError):
        sim.set_optic_fiber_layer(y_range='invalid')

    x_range = (-100, 100)
    y_range = (0, 200)
    sim.set_optic_fiber_layer(x_range=x_range, y_range=y_range,
                              save_data=False)
    npt.assert_equal(sim.ofl.gridx.min(), x_range[0])
    npt.assert_equal(sim.ofl.gridx.max(), x_range[1])
    npt.assert_equal(sim.ofl.gridy.min(), y_range[0])
    npt.assert_equal(sim.ofl.gridy.max(), y_range[1])
    npt.assert_equal(sim.ofl.x_range, x_range)
    npt.assert_equal(sim.ofl.y_range, y_range)

    # Smoke test
    implant = implants.ElectrodeArray('epiretinal', 10, 0, 0, 0)
    sim = p2p.Simulation(implant, engine='serial')
    sim.set_optic_fiber_layer(x_range=0, y_range=0)
    sim.set_optic_fiber_layer(x_range=[0, 0], y_range=[0, 0])
    sim.set_optic_fiber_layer()


def test_Simulation_set_ganglion_cell_layer():
    # A valid ganglion cell model
    class Valid(p2p.retina.BaseModel):

        def model_cascade(self, inval):
            return inval

    # Smoke test custom model
    implant = implants.ElectrodeArray('epiretinal', 10, 0, 0, 0)
    sim = p2p.Simulation(implant, engine='serial')
    sim.set_optic_fiber_layer(x_range=0, y_range=0)

    valid_model = Valid(tsample=0.2)
    sim.set_ganglion_cell_layer(valid_model)
    npt.assert_equal(isinstance(sim.gcl, p2p.retina.BaseModel), True)
    npt.assert_equal(sim.gcl.tsample, 0.2)

    # Smoke test latest model
    for modelstr in ['latest', 'Latest', 'LATEST']:
        sim.set_ganglion_cell_layer(modelstr, lweight=42)
        npt.assert_equal(isinstance(sim.gcl, p2p.retina.TemporalModel), True)
        npt.assert_equal(sim.gcl.lweight, 42)
        sim.set_ganglion_cell_layer(modelstr, unknown_param=2)  # smoke

    # Smoke test Nanduri model
    for modelstr in ['Nanduri2012', 'nanduri2012', 'NANDURI2012']:
        sim.set_ganglion_cell_layer(modelstr, tau3=42)
        npt.assert_equal(isinstance(sim.gcl, p2p.retina.Nanduri2012), True)
        npt.assert_equal(sim.gcl.tau3, 42)
        sim.set_ganglion_cell_layer(modelstr, unknown_param=2)  # smoke

    # Smoke test Horsager model
    for modelstr in ['Horsager2009', 'horsager', 'HORSAGER2009']:
        sim.set_ganglion_cell_layer(modelstr, tau3=42)
        npt.assert_equal(isinstance(sim.gcl, p2p.retina.Horsager2009), True)
        npt.assert_equal(sim.gcl.tau3, 42)
        sim.set_ganglion_cell_layer(modelstr, unknown_param=2)  # smoke

    # Model unknown
    with pytest.raises(ValueError):
        sim.set_ganglion_cell_layer('unknown-model')
    with pytest.raises(ValueError):
        sim.set_ganglion_cell_layer(implants.ArgusII())


def test_get_brightest_frame():
    # Pick a few raindom frames and place brightest pixel therein
    num_frames = 10
    rand_idx = np.random.randint(0, num_frames, 5)
    for idx in rand_idx:
        # Set the brightes pixel in frame `idx` of a random vector
        tsdata = np.random.rand(2, 2, num_frames)
        tsdata[1, 1, idx] = 2.0

        # Make sure function returns the right frame
        ts = utils.TimeSeries(1, tsdata)
        brightest = p2p.get_brightest_frame(ts)
        npt.assert_equal(brightest.data.max(), tsdata.max())
        npt.assert_equal(brightest.data, tsdata[:, :, idx])



# -*implants -*-
"""

Functions for creating retinal implants

"""
import numpy as np
import six
import copy
import logging

from pulse2percept import utils
from pulse2percept import implants
from pulse2percept import files

# Rather than trying to import these all over, try once and then remember
# by setting a flag.
try:
    import skimage.io as sio
    import skimage.transform as sit
    import skimage.color as sic
    has_skimage = True
except (ImportError, AttributeError):
    # Might also raise "dict object has no attribute 'io'"
    has_skimage = False


class MonophasicPulse(utils.TimeSeries):

    def __init__(self, ptype, pdur, tsample, delay_dur=0, stim_dur=None):
        """A pulse with a single phase

        Parameters
        ----------
        ptype : {'anodic', 'cathodic'}
            Pulse type. Anodic pulses have positive current amplitude,
            cathodic pulses have negative amplitude.
        pdur : float
            Pulse duration (s).
        tsample : float
            Sampling time step (s).
        delay_dur : float, optional
            Pulse delay (s). Pulse will be zero-padded (prepended) to deliver
            the pulse only after `delay_dur` milliseconds. Default: 0.
        stim_dur : float, optional
            Stimulus duration (ms). Pulse will be zero-padded (appended) to fit
            the stimulus duration. Default: No additional zero padding,
            `stim_dur` is `pdur`+`delay_dur`.
        """
        if tsample <= 0:
            raise ValueError("tsample must be a non-negative float.")

        if stim_dur is None:
            stim_dur = pdur + delay_dur

        # Convert durations to number of samples
        pulse_size = int(np.round(pdur / tsample))
        delay_size = int(np.round(delay_dur / tsample))
        stim_size = int(np.round(stim_dur / tsample))

        if ptype == 'cathodic':
            pulse = -np.ones(pulse_size)
        elif ptype == 'anodic':
            pulse = np.ones(pulse_size)
        else:
            raise ValueError("Acceptable values for `ptype` are 'anodic', "
                             "'cathodic'.")

        pulse = np.concatenate((np.zeros(delay_size), pulse,
                                np.zeros(stim_size)))
        utils.TimeSeries.__init__(self, tsample, pulse[:stim_size])


class BiphasicPulse(utils.TimeSeries):

    def __init__(self, ptype, pdur, tsample, interphase_dur=0):
        """A charge-balanced pulse with a cathodic and anodic phase

        A single biphasic pulse with duration `pdur` per phase,
        separated by `interphase_dur` is returned.

        Parameters
        ----------
        ptype : {'cathodicfirst', 'anodicfirst'}
            A cathodic-first pulse has the negative phase first, whereas an
            anodic-first pulse has the positive phase first.
        pdur : float
            Duration of single (positive or negative) pulse phase in seconds.
        tsample : float
            Sampling time step in seconds.
        interphase_dur : float, optional
            Duration of inter-phase interval (between positive and negative
            pulse) in seconds. Default: 0.
        """
        if tsample <= 0:
            raise ValueError("tsample must be a non-negative float.")

        # Get the two monophasic pulses
        on = MonophasicPulse('anodic', pdur, tsample, 0, pdur)
        off = MonophasicPulse('cathodic', pdur, tsample, 0, pdur)

        # Insert interphase gap if necessary
        gap = np.zeros(int(round(interphase_dur / tsample)))

        # Order the pulses
        if ptype == 'cathodicfirst':
            # has negative current first
            pulse = np.concatenate((off.data, gap), axis=0)
            pulse = np.concatenate((pulse, on.data), axis=0)
        elif ptype == 'anodicfirst':
            pulse = np.concatenate((on.data, gap), axis=0)
            pulse = np.concatenate((pulse, off.data), axis=0)
        else:
            raise ValueError("Acceptable values for `type` are "
                             "'anodicfirst' or 'cathodicfirst'")
        utils.TimeSeries.__init__(self, tsample, pulse)


class PulseTrain(utils.TimeSeries):

    def __init__(self, tsample, freq=20, amp=20, dur=0.5, delay=0,
                 pulse_dur=0.45 / 1000, interphase_dur=0.45 / 1000,
                 pulsetype='cathodicfirst',
                 pulseorder='pulsefirst'):
        """A train of biphasic pulses

        Parameters
        ----------
        tsample : float
            Sampling time step (seconds).
        freq : float, optional, default: 20 Hz
            Frequency of the pulse envelope (Hz).
        amp : float, optional, default: 20 uA
            Max amplitude of the pulse train in micro-amps.
        dur : float, optional, default: 0.5 seconds
            Stimulus duration in seconds.
        delay : float, optional, default: 0
            Delay until stimulus on-set in seconds.
        pulse_dur : float, optional, default: 0.45 ms
            Single-pulse duration in seconds.
        interphase_duration : float, optional, default: 0.45 ms
            Single-pulse interphase duration (the time between the positive
            and negative phase) in seconds.
        pulsetype : str, optional, default: 'cathodicfirst'
            Pulse type {'cathodicfirst' | 'anodicfirst'}, where
            'cathodicfirst' has the negative phase first.
        pulseorder : str, optional, default: 'pulsefirst'
            Pulse order {'gapfirst' | 'pulsefirst'}, where
            'pulsefirst' has the pulse first, followed by the gap.
            'gapfirst' has it the other way round.
        """
        if tsample <= 0:
            raise ValueError("tsample must be a non-negative float.")

        # Stimulus size given by `dur`
        stim_size = int(np.round(float(dur) / tsample))

        # Make sure input is non-trivial, else return all zeros
        if np.isclose(freq, 0) or np.isclose(amp, 0):
            utils.TimeSeries.__init__(self, tsample, np.zeros(stim_size))
            return

        # Envelope size (single pulse + gap) given by `freq`
        # Note that this can be larger than `stim_size`, but we will trim
        # the stimulus to proper length at the very end.
        envelope_size = int(np.round(1.0 / float(freq) / tsample))
        if envelope_size > stim_size:
            debug_s = ("Envelope size (%d) clipped to "
                       "stimulus size (%d) for freq=%f" % (envelope_size,
                                                           stim_size,
                                                           freq))
            logging.getLogger(__name__).debug(debug_s)
            envelope_size = stim_size

        # Delay given by `delay`
        delay_size = int(np.round(float(delay) / tsample))

        if delay_size < 0:
            raise ValueError("Delay cannot be negative.")
        delay = np.zeros(delay_size)

        # Single pulse given by `pulse_dur`
        pulse = amp * BiphasicPulse(pulsetype, pulse_dur, tsample,
                                    interphase_dur).data
        pulse_size = pulse.size
        if pulse_size < 0:
            raise ValueError("Single pulse must fit within 1/freq interval.")

        # Then gap is used to fill up what's left
        gap_size = envelope_size - (delay_size + pulse_size)
        if gap_size < 0:
            logging.error("Envelope (%d) can't fit pulse (%d) + delay (%d)" %
                          (envelope_size, pulse_size, delay_size))
            raise ValueError("Pulse and delay must fit within 1/freq "
                             "interval.")
        gap = np.zeros(gap_size)

        pulse_train = np.array([])
        for j in range(int(np.ceil(dur * freq))):
            if pulseorder == 'pulsefirst':
                pulse_train = np.concatenate((pulse_train, delay, pulse,
                                              gap), axis=0)
            elif pulseorder == 'gapfirst':
                pulse_train = np.concatenate((pulse_train, delay, gap,
                                              pulse), axis=0)
            else:
                raise ValueError("Acceptable values for `pulseorder` are "
                                 "'pulsefirst' or 'gapfirst'")

        # If `freq` is not a nice number, the resulting pulse train might not
        # have the desired length
        if pulse_train.size < stim_size:
            fill_size = stim_size - pulse_train.shape[-1]
            pulse_train = np.concatenate((pulse_train, np.zeros(fill_size)),
                                         axis=0)

        # Trim to correct length (takes care of too long arrays, too)
        pulse_train = pulse_train[:stim_size]

        utils.TimeSeries.__init__(self, tsample, pulse_train)


def image2pulsetrain(img, implant, coding='amplitude', valrange=[0, 50],
                     max_contrast=False, const_val=20, invert=False,
                     tsample=0.005 / 1000, dur=0.5, pulsedur=0.5 / 1000.,
                     interphasedur=0.5 / 1000., pulsetype='cathodicfirst'):
    """Converts an image into a series of pulse trains

    This function creates an input stimulus from an RGB or grayscale image.
    The image is down-sampled to fit the spatial layout of the implant
    (currently supported are ArgusI and ArgusII arrays).
    Requires Scikit-Image.

    Parameters
    ----------
    img : str|array_like
        An input image, either a valid filename (string) or a numpy array
        (row x col x channels).
    implant : p2p.implants.ElectrodeArray
        An ElectrodeArray object that describes the implant.
    coding : {'amplitude', 'frequency'}, optional
        A string describing the coding scheme:
        - 'amplitude': Image intensity is linearly converted to a current
                       amplitude between `valrange[0]` and `valrange[1]`.
                       Frequency is held constant at `const_freq`.
        - 'frequency': Image intensity is linearly converted to a pulse
                       frequency between `valrange[0]` and `valrange[1]`.
                       Amplitude is held constant at `const_amp`.
        Default: 'amplitude'
    valrange : list, optional
        Range of stimulation values to be used (If `coding` is 'amplitude',
        specifies min and max current; if `coding` is 'frequency', specifies
        min and max frequency).
        Default: [0, 50]
    max_contrast : bool, optional
        Flag wether to maximize image contrast (True) or not (False).
        Default: False
    const_val : float, optional
        For frequency coding: The constant amplitude value to be used for all
        pulse trains. For amplitude coding: The constant frequency value to
        be used for all pulse trains.
        Default: 20
    invert : bool, optional
        Flag whether to invert the grayscale values of the image (True) or
        not (False).
        Default: False
    tsample : float, optional
        Sampling time step (seconds). Default: 0.005 / 1000 seconds.
    dur : float, optional
        Stimulus duration (seconds). Default: 0.5 seconds.
    pulsedur : float, optional
        Duration of single (positive or negative) pulse phase in seconds.
    interphasedur : float, optional
        Duration of inter-phase interval (between positive and negative
        pulse) in seconds.
    pulsetype : {'cathodicfirst', 'anodicfirst'}, optional
        A cathodic-first pulse has the negative phase first, whereas an
        anodic-first pulse has the positive phase first.

    Returns
    -------
    pulses : list
        A list of p2p.stimuli.PulseTrain objects, one for each electrode in
        the implant.

    """
    if not has_skimage:
        # We don't want to repeatedly import Scikit-Image. This would (e.g.)
        # unnecessarily slow down `video2pulsetrain`.
        raise ImportError("You do not have scikit-image installed. "
                          "You can install it via $ pip install scikit-image.")

    # Make sure range of values is valid
    assert len(valrange) == 2 and valrange[1] > valrange[0]

    isargus1 = isinstance(implant, implants.ArgusI)
    isargus2 = isinstance(implant, implants.ArgusII)
    if not isargus1 and not isargus2:
        raise TypeError("For now, implant must be of type implants.ArgusI or "
                        "implants.ArgusII.")

    if isinstance(img, six.string_types):
        # Load image from filename
        img_orig = sio.imread(img, as_grey=True).astype(np.float32)
        logging.getLogger(__name__).info("Loaded file '%s'." % img)
    else:
        if img.ndim == 2:
            # Grayscale
            img_orig = img.astype(np.float32)
        else:
            # Assume RGB, convert to grayscale
            assert img.shape[-1] == 3
            img_orig = sic.rgb2gray(np.array(img)).astype(np.float32)

    # Make sure all pixels are between 0 and 1
    if img_orig.max() > 1.0:
        img_orig /= 255.0

    # Let Scikit-Image do the resampling: Downscale image to fit array
    # Use mode 'reflect' for np.pad: Pads with the reflection of the vector
    # mirrored on the first and last values of the vector along each axis.
    if isargus1:
        img_stim = sit.resize(img_orig, (4, 4), mode='reflect')
    elif isargus2:
        img_stim = sit.resize(img_orig, (6, 10), mode='reflect')

    # If specified, invert the mapping of grayscale values:
    if invert:
        img_stim = 1.0 - img_stim

    # If specified, maximize the contrast in the image:
    if max_contrast:
        img_stim -= img_stim.min()
        if img_stim.max() > 0:
            img_stim /= img_stim.max()

    # With all pixels between 0 and 1, now scale to valrange
    assert np.all(img_stim >= 0.0) and np.all(img_stim <= 1.0)
    img_stim = img_stim * np.diff(valrange) + valrange[0]
    assert np.all(img_stim >= valrange[0]) and np.all(img_stim <= valrange[1])

    stim = []
    for _, px in np.ndenumerate(img_stim):
        if coding == 'amplitude':
            amp = px
            freq = const_val
        elif coding == 'frequency':
            amp = const_val
            freq = px
        else:
            e_s = "Acceptable values for `coding` are 'amplitude' or"
            e_s += "'frequency'."
            raise ValueError(e_s)

        pt = PulseTrain(tsample, freq=freq, amp=amp, dur=dur,
                        pulse_dur=pulsedur,
                        interphase_dur=interphasedur,
                        pulsetype=pulsetype)
        stim.append(pt)

    return stim


def video2pulsetrain(filename, implant, framerate=20,
                     coding='amplitude', valrange=[0, 50],
                     max_contrast=False, const_val=20, invert=False,
                     tsample=0.005 / 1000, pulsedur=0.5 / 1000.,
                     interphasedur=0.5 / 1000., pulsetype='cathodicfirst',
                     ffmpeg_path=None, libav_path=None):
    """Converts a video into a series of pulse trains

    This function creates an input stimulus from a video.
    Every frame of the video is passed to `image2pulsetrain`, where it is
    down-sampled to fit the spatial layout of the implant (currently supported
    are ArgusI and ArgusII arrays).
    In this mapping, rows of the image correspond to rows in the implant
    (top row, Argus I: A1 B1 C1 D1, Argus II: A1 A2 ... A10).

    Requires Scikit-Image and Scikit-Video.

    Parameters
    ----------
    img : str|array_like
        An input image, either a valid filename (string) or a numpy array
        (row x col x channels).
    implant : p2p.implants.ElectrodeArray
        An ElectrodeArray object that describes the implant.
    coding : {'amplitude', 'frequency'}, optional
        A string describing the coding scheme:
        - 'amplitude': Image intensity is linearly converted to a current
                       amplitude between `valrange[0]` and `valrange[1]`.
                       Frequency is held constant at `const_freq`.
        - 'frequency': Image intensity is linearly converted to a pulse
                       frequency between `valrange[0]` and `valrange[1]`.
                       Amplitude is held constant at `const_amp`.
        Default: 'amplitude'
    valrange : list, optional
        Range of stimulation values to be used (If `coding` is 'amplitude',
        specifies min and max current; if `coding` is 'frequency', specifies
        min and max frequency).
        Default: [0, 50]
    max_contrast : bool, optional
        Flag wether to maximize image contrast (True) or not (False).
        Default: False
    const_val : float, optional
        For frequency coding: The constant amplitude value to be used for all
        pulse trains. For amplitude coding: The constant frequency value to
        be used for all pulse trains.
        Default: 20
    invert : bool, optional
        Flag whether to invert the grayscale values of the image (True) or
        not (False).
        Default: False
    tsample : float, optional
        Sampling time step (seconds). Default: 0.005 / 1000 seconds.
    dur : float, optional
        Stimulus duration (seconds). Default: 0.5 seconds.
    pulsedur : float, optional
        Duration of single (positive or negative) pulse phase in seconds.
    interphasedur : float, optional
        Duration of inter-phase interval (between positive and negative
        pulse) in seconds.
    pulsetype : {'cathodicfirst', 'anodicfirst'}, optional
        A cathodic-first pulse has the negative phase first, whereas an
        anodic-first pulse has the positive phase first.

    Returns
    -------
    pulses : list
        A list of p2p.stimuli.PulseTrain objects, one for each electrode in
        the implant.

    """

    # Load generator to read video frame-by-frame
    reader = files.load_video_generator(filename, ffmpeg_path, libav_path)

    # Temporarily increase logger level to suppress info messages
    current_level = logging.getLogger(__name__).getEffectiveLevel()
    logging.getLogger(__name__).setLevel(logging.WARN)

    # Convert the desired framerate to a duration (seconds)
    dur = 1.0 / framerate

    # Read one frame at a time, and append to previous frames
    video = []
    for img in reader.nextFrame():
        frame = image2pulsetrain(img, implant, coding=coding,
                                 valrange=valrange, max_contrast=max_contrast,
                                 const_val=const_val, invert=invert,
                                 tsample=tsample, dur=dur, pulsedur=pulsedur,
                                 interphasedur=interphasedur,
                                 pulsetype=pulsetype)
        if video:
            # List of pulse trains: Append new frame to each element
            [v.append(f) for v, f in zip(video, frame)]
        else:
            # Initialize with a list of pulse trains
            video = frame

    # Restore logger level
    logging.getLogger(__name__).setLevel(current_level)

    return video


def parse_pulse_trains(stim, implant):
    """Parse input stimulus and convert to list of pulse trains

    Parameters
    ----------
    stim : utils.TimeSeries|list|dict
        There are several ways to specify an input stimulus:

        - For a single-electrode array, pass a single pulse train; i.e., a
          single utils.TimeSeries object.
        - For a multi-electrode array, pass a list of pulse trains, where
          every pulse train is a utils.TimeSeries object; i.e., one pulse
          train per electrode.
        - For a multi-electrode array, specify all electrodes that should
          receive non-zero pulse trains by name in a dictionary. The key
          of each element is the electrode name, the value is a pulse train.
          Example: stim = {'E1': pt, 'stim': pt}, where 'E1' and 'stim' are
          electrode names, and `pt` is a utils.TimeSeries object.
    implant : p2p.implants.ElectrodeArray
        A p2p.implants.ElectrodeArray object that describes the implant.

    Returns
    -------
    A list of pulse trains; one pulse train per electrode.
    """
    # Parse input stimulus
    if isinstance(stim, utils.TimeSeries):
        # `stim` is a single object: This is only allowed if the implant
        # has only one electrode
        if implant.num_electrodes > 1:
            e_s = "More than 1 electrode given, use a list of pulse trains"
            raise ValueError(e_s)
        pt = [copy.deepcopy(stim)]
    elif isinstance(stim, dict):
        # `stim` is a dictionary: Look up electrode names and assign pulse
        # trains, fill the rest with zeros

        # Get right size from first dict element, then generate all zeros
        idx0 = list(stim.keys())[0]
        pt_zero = utils.TimeSeries(stim[idx0].tsample,
                                   np.zeros_like(stim[idx0].data))
        pt = [pt_zero] * implant.num_electrodes

        # Iterate over dictionary and assign non-zero pulse trains to
        # corresponding electrodes
        for key, value in stim.items():
            el_idx = implant.get_index(key)
            if el_idx is not None:
                pt[el_idx] = copy.deepcopy(value)
            else:
                e_s = "Could not find electrode with name '%s'" % key
                raise ValueError(e_s)
    else:
        # Else, `stim` must be a list of pulse trains, one for each electrode
        if len(stim) != implant.num_electrodes:
            e_s = "Number of pulse trains must match number of electrodes"
            raise ValueError(e_s)
        pt = copy.deepcopy(stim)

    return pt



import numpy as np
import numpy.testing as npt
import pytest
try:
    # Python 3
    from unittest import mock
except ImportError:
    # Python 2
    import mock

try:
    # Python 3
    from imp import reload
except ImportError:
    pass

from .. import stimuli
from .. import implants
from .. import utils


def test_MonophasicPulse():
    tsample = 1.0

    for pdur in range(10):
        for ddur in range(10):
            for sdur in [None, 5, 10, 100]:
                for ptype in ['anodic', 'cathodic']:
                    pulse = stimuli.MonophasicPulse(ptype, pdur,
                                                    tsample, ddur,
                                                    sdur)

                    # Make sure the pulse length is correct:
                    # When stimulus duration is not specified, stimulus must
                    # tightly fit pulse + delay. When stimulus direction is
                    # specified, pulse duration must match that number exactly.
                    if sdur is None:
                        npt.assert_equal(pulse.data.size, pdur + ddur)
                    else:
                        npt.assert_equal(pulse.data.size, sdur)

                    # Make sure delay is correct
                    delay = pulse.data[:ddur]
                    npt.assert_equal(np.allclose(delay, 0.0), True)

                    # Make sure pulse length and amplitude are correct
                    if pdur > 0:
                        # Actually, depending on settings, pulse duration might
                        # be different from what was specified:
                        if sdur is not None and pdur + ddur > sdur:
                            # Stim is trimmed, adjust pdur
                            actual_pdur = np.maximum(0, sdur - ddur)
                        else:
                            actual_pdur = pdur

                        # Find maximma/minima
                        idx_min = np.isclose(pulse.data, -1.0)
                        idx_max = np.isclose(pulse.data, 1.0)
                        if ptype == 'anodic':
                            npt.assert_equal(np.sum(idx_max), actual_pdur)
                            if actual_pdur > 0:
                                npt.assert_equal(pulse.data.max(), 1.0)
                        else:
                            npt.assert_equal(np.sum(idx_min), actual_pdur)
                            if actual_pdur > 0:
                                npt.assert_equal(pulse.data.min(), -1.0)

    # Invalid pulsetype
    with pytest.raises(ValueError):
        stimuli.MonophasicPulse('anodicfirst', 10, 0.1)
    with pytest.raises(ValueError):
        stimuli.MonophasicPulse('cathodicfirst', 10, 0.1)


def test_BiphasicPulse():
    for ptype in ['cathodicfirst', 'anodicfirst']:
        for pdur in [0.25 / 1000, 0.45 / 1000, 0.65 / 1000]:
            for interphase_dur in [0, 0.25 / 1000, 0.45 / 1000, 0.65 / 1000]:
                for tsample in [5e-6, 1e-5, 5e-5]:
                    # generate pulse
                    pulse = stimuli.BiphasicPulse(ptype, pdur,
                                                  tsample,
                                                  interphase_dur)

                    # predicted length
                    pulse_gap_dur = 2 * pdur + interphase_dur

                    # make sure length is correct
                    npt.assert_equal(pulse.shape[-1],
                                     int(np.round(pulse_gap_dur /
                                                  tsample)))

                    # make sure amplitude is correct: negative peak,
                    # zero (in case of nonnegative interphase dur),
                    # positive peak
                    if interphase_dur > 0:
                        npt.assert_equal(np.unique(pulse.data),
                                         np.array([-1, 0, 1]))
                    else:
                        npt.assert_equal(np.unique(pulse.data),
                                         np.array([-1, 1]))

                    # make sure pulse order is correct
                    idx_min = np.where(pulse.data == pulse.data.min())
                    idx_max = np.where(pulse.data == pulse.data.max())
                    if ptype == 'cathodicfirst':
                        # cathodicfirst should have min first
                        npt.assert_equal(idx_min[0] < idx_max[0], True)
                    else:
                        npt.assert_equal(idx_min[0] < idx_max[0], False)

    # Invalid pulsetype
    with pytest.raises(ValueError):
        stimuli.BiphasicPulse('anodic', 10, 0.1)
    with pytest.raises(ValueError):
        stimuli.BiphasicPulse('cathodic', 10, 0.1)


def test_PulseTrain():
    dur = 0.5
    pdur = 0.45 / 1000
    tsample = 5e-6
    ampl = 20.0
    freq = 5.0

    # First an easy one (sawtooth)...
    for scale in [1.0, 2.0, 5.0, 10.0]:
        pt = stimuli.PulseTrain(tsample=0.1 * scale,
                                dur=1.0 * scale, freq=freq / scale,
                                amp=ampl * scale,
                                pulse_dur=0.1 * scale,
                                interphase_dur=0,
                                pulsetype='cathodicfirst',
                                pulseorder='pulsefirst')
        npt.assert_equal(np.sum(np.isclose(pt.data, ampl * scale)), freq)
        npt.assert_equal(np.sum(np.isclose(pt.data, -ampl * scale)), freq)
        npt.assert_equal(pt.data[0], -ampl * scale)
        npt.assert_equal(pt.data[-1], ampl * scale)
        npt.assert_equal(len(pt.data), 10)

    # Then some more general ones...
    # Size of array given stimulus duration
    stim_size = int(np.round(dur / tsample))

    # All empty pulse trains: Expect no division by zero errors
    for amp in [0, 20]:
        p2pt = stimuli.PulseTrain(freq=0, amp=ampl, dur=dur,
                                  pulse_dur=pdur,
                                  interphase_dur=pdur,
                                  tsample=tsample)
        npt.assert_equal(p2pt.data, np.zeros(stim_size))

    # Non-zero pulse trains: Expect right length, pulse order, etc.
    for freq in [8, 13.8, 20]:
        for pulsetype in ['cathodicfirst', 'anodicfirst']:
            for delay in [0, 10 / 1000]:
                for pulseorder in ['pulsefirst', 'gapfirst']:
                    p2pt = stimuli.PulseTrain(freq=freq,
                                              dur=dur,
                                              pulse_dur=pdur,
                                              interphase_dur=pdur,
                                              delay=delay,
                                              tsample=tsample,
                                              amp=ampl,
                                              pulsetype=pulsetype,
                                              pulseorder=pulseorder)

                    # make sure length is correct
                    npt.assert_equal(p2pt.data.size, stim_size)

                    # make sure amplitude is correct
                    npt.assert_equal(np.unique(p2pt.data),
                                     np.array([-ampl, 0, ampl]))

                    # make sure pulse type is correct
                    idx_min = np.where(p2pt.data == p2pt.data.min())
                    idx_max = np.where(p2pt.data == p2pt.data.max())
                    if pulsetype == 'cathodicfirst':
                        # cathodicfirst should have min first
                        npt.assert_equal(idx_min[0] < idx_max[0], True)
                    else:
                        npt.assert_equal(idx_min[0] < idx_max[0], False)

                    # Make sure frequency is correct
                    # Need to trim size if `freq` is not a nice number
                    envelope_size = int(np.round(1.0 / float(freq) / tsample))
                    single_pulse_dur = int(np.round(1.0 * pdur / tsample))
                    num_pulses = int(np.floor(dur * freq))  # round down
                    trim_sz = envelope_size * num_pulses
                    idx_min = np.where(p2pt.data[:trim_sz] == p2pt.data.min())
                    idx_max = np.where(p2pt.data[:trim_sz] == p2pt.data.max())
                    npt.assert_equal(idx_max[0].shape[-1],
                                     num_pulses * single_pulse_dur)
                    npt.assert_equal(idx_min[0].shape[-1],
                                     num_pulses * single_pulse_dur)

                    # make sure pulse order is correct
                    delay_dur = int(np.round(delay / tsample))
                    envelope_dur = int(np.round((1 / freq) / tsample))
                    if pulsetype == 'cathodicfirst':
                        val = p2pt.data.min()  # expect min first
                    else:
                        val = p2pt.data.max()  # expect max first
                    if pulseorder == 'pulsefirst':
                        idx0 = delay_dur  # expect pulse first, then gap
                    else:
                        idx0 = envelope_dur - 3 * single_pulse_dur
                    npt.assert_equal(p2pt.data[idx0], val)
                    npt.assert_equal(p2pt.data[idx0 + envelope_dur], val)

    # Invalid values
    with pytest.raises(ValueError):
        stimuli.PulseTrain(0.1, delay=-10)
    with pytest.raises(ValueError):
        stimuli.PulseTrain(0.1, pulse_dur=-10)
    with pytest.raises(ValueError):
        stimuli.PulseTrain(0.1, freq=1000, pulse_dur=10)
    with pytest.raises(ValueError):
        stimuli.PulseTrain(0.1, pulseorder='cathodicfirst')
    with pytest.raises(ValueError):
        stimuli.PulseTrain(0)

    # Smoke test envelope_size > stim_size
    stimuli.PulseTrain(1, freq=0.01, dur=0.01)


def test_image2pulsetrain():
    # Range of values
    amp_min = 2
    amp_max = 15

    # Create a standard Argus I array
    implant = implants.ArgusI()

    # Create a small image with 1 pixel per electrode
    img = np.zeros((4, 4))

    # An all-zero image should give us a really boring stimulation protocol
    pulses = stimuli.image2pulsetrain(img, implant, valrange=[amp_min,
                                                              amp_max])
    for pt in pulses:
        npt.assert_equal(pt.data.max(), amp_min)

    # Now put some structure in the image
    img[1, 1] = img[1, 2] = img[2, 1] = img[2, 2] = 0.75

    expected_max = [amp_max, 0.75 * (amp_max - amp_min) + amp_min]
    for max_contrast, val_max in zip([True, False], expected_max):
        pt = stimuli.image2pulsetrain(img, implant, coding='amplitude',
                                      max_contrast=max_contrast,
                                      valrange=[amp_min, amp_max])

        # Make sure we have one pulse train per electrode
        npt.assert_equal(len(pt), implant.num_electrodes)

        # Make sure the brightest electrode has `amp_max`
        npt.assert_almost_equal(np.max([p.data.max() for p in pt]),
                                val_max)

        # Make sure the dimmest electrode has `amp_min` as max amplitude
        npt.assert_almost_equal(np.min([np.abs(p.data).max() for p in pt]),
                                amp_min)

    # Invalid implant
    with pytest.raises(TypeError):
        stimuli.image2pulsetrain("rainbow_cat.jpg", np.zeros(10))
    with pytest.raises(TypeError):
        e_array = implants.ElectrodeArray('epiretinal', 100, 0, 0)
        stimuli.image2pulsetrain("rainbow_cat.jpg", e_array)

    # Invalid image
    with pytest.raises(IOError):
        stimuli.image2pulsetrain("rainbow_cat.jpg", implants.ArgusI())

    # Smoke-test RGB
    stimuli.image2pulsetrain(np.zeros((10, 10, 3)), implants.ArgusI())

    # Smoke-test invert
    stimuli.image2pulsetrain(np.zeros((10, 10, 3)), implants.ArgusI(),
                             invert=True)

    # Smoke-test normalize
    stimuli.image2pulsetrain(np.ones((10, 10, 3)) * 2,
                             implants.ArgusI(), invert=True)

    # Smoke-test frequency coding
    stimuli.image2pulsetrain(np.zeros((10, 10, 3)), implants.ArgusI(),
                             coding='frequency')

    # Invalid coding
    with pytest.raises(ValueError):
        stimuli.image2pulsetrain(np.zeros((10, 10)), implants.ArgusI(),
                                 coding='n/a')

    # Trigger an import error
    with mock.patch.dict("sys.modules", {"skimage": {}, "skimage.io": {}}):
        with pytest.raises(ImportError):
            reload(stimuli)
            stimuli.image2pulsetrain(img, implant)


def test_video2pulsetrain():
    reload(stimuli)
    implant = implants.ArgusI()

    with pytest.raises(OSError):
        stimuli.video2pulsetrain('no-such-file.avi', implant)

    # Smoke-test example video
    from skvideo import datasets
    stimuli.video2pulsetrain(datasets.bikes(), implant)


def test_parse_pulse_trains():
    # Specify pulse trains in a number of different ways and make sure they
    # are all identical after parsing

    # Create some p2p.implants
    argus = implants.ArgusI()
    simple = implants.ElectrodeArray('subretinal', 0, 0, 0, 0)

    pt_zero = utils.TimeSeries(1, np.zeros(1000))
    pt_nonzero = utils.TimeSeries(1, np.random.rand(1000))

    # Test 1
    # ------
    # Specify wrong number of pulse trains
    with pytest.raises(ValueError):
        stimuli.parse_pulse_trains(pt_nonzero, argus)
    with pytest.raises(ValueError):
        stimuli.parse_pulse_trains([pt_nonzero], argus)
    with pytest.raises(ValueError):
        stimuli.parse_pulse_trains([pt_nonzero] *
                                   (argus.num_electrodes - 1),
                                   argus)
    with pytest.raises(ValueError):
        stimuli.parse_pulse_trains([pt_nonzero] * 2, simple)

    # Test 2
    # ------
    # Send non-zero pulse train to specific electrode
    el_name = 'B3'
    el_idx = argus.get_index(el_name)

    # Specify a list of 16 pulse trains (one for each electrode)
    pt0_in = [pt_zero] * argus.num_electrodes
    pt0_in[el_idx] = pt_nonzero
    pt0_out = stimuli.parse_pulse_trains(pt0_in, argus)

    # Specify a dict with non-zero pulse trains
    pt1_in = {el_name: pt_nonzero}
    pt1_out = stimuli.parse_pulse_trains(pt1_in, argus)

    # Make sure the two give the same result
    for p0, p1 in zip(pt0_out, pt1_out):
        npt.assert_equal(p0.data, p1.data)

    # Test 3
    # ------
    # Smoke testing
    stimuli.parse_pulse_trains([pt_zero] * argus.num_electrodes, argus)
    stimuli.parse_pulse_trains(pt_zero, simple)
    stimuli.parse_pulse_trains([pt_zero], simple)
